"""Functions helpful for demonstrating the use of BLAST-Lite in Python."""

import numpy as np
import pandas as pd

from blast.utils.functions import assemble_one_year_input

def generate_sample_data(kind: str = "synthetic") -> dict:
    """
    Generate synthetic sample data for demonstration purposes.
    Options:

        1. Completely synthetic data
        2. Data from a small personal EV in Honolulu, Hawaii.
        3. Data from a large personal EV in Honolulu, Hawaii.
        4. Data from a commercial EV in Honolulu, Hawaii.

    Args:
        kind (str): One of 'synthetic', 'ev_smallbattery', 'ev_largebattery',
        'ev_commercial', 'ev_commercial_lowdod', 'ev_commercial_lowdod_lowsoc'

    Returns:
        dict:  Dictionary with keys {'Time_s', 'SOC', 'Temperature_C'}
    """

    _allowed_kinds = {
        "synthetic",
        "ev_smallbattery",
        "ev_largebattery",
        "ev_commercial",
        "ev_commercial_lowdod",
        "ev_commercial_lowdod_lowsoc",
    }

    if kind == "synthetic":
        # Set up example time
        years = 10
        hours = years * 365 * 24
        t_hours = np.arange(hours + 1)

        # Set up an example profile
        soc = np.tile(
            np.array(
                [
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    0.4,
                    0.4,
                    0.6,
                    0.6,
                    1,
                    1,
                    0.9,
                    0.8,
                    0.7,
                    0.6,
                    0.5,
                    0.4,
                    0.3,
                    0.2,
                    0.2,
                    0.2,
                    0.2,
                    0.6,
                ]
            ),
            years * 365,
        )
        soc = np.append(soc, 1)

        # Set up example temperature
        TdegC = np.tile(
            np.concatenate([np.linspace(25, 35, 12), np.linspace(35, 25, 12)]),
            int(len(t_hours) / 24),
        ) + np.tile(
            np.concatenate(
                [np.linspace(-5, 5, 24 * 182), np.linspace(5, -5, 24 * 183)]
            ),
            int(len(t_hours) / (24 * 365)),
        )
        TdegC = np.append(TdegC, 20)

        t_secs = t_hours * 3600

        input = {"Time_s": t_secs, "SOC": soc, "Temperature_C": TdegC}

    elif kind == "ev_smallbattery":
        climate = pd.read_csv("examples/climates/nsrdb_honolulu.csv")
        ev_smallbattery = pd.read_csv(
            "examples/application profiles/personal_ev_smallbatt.csv"
        )
        ev_smallbattery = ev_smallbattery.iloc[
            np.linspace(0, 24 * 3600 * 7 - 1, 24 * 7)
        ]
        input = assemble_one_year_input(ev_smallbattery, climate)

    elif kind == "ev_largebattery":
        climate = pd.read_csv("examples/climates/nsrdb_honolulu.csv")
        ev_largebattery = pd.read_csv(
            "examples/application profiles/personal_ev_largebatt.csv"
        )
        ev_largebattery = ev_largebattery.iloc[
            np.linspace(0, 24 * 3600 * 7 - 1, 24 * 7)
        ]
        input = assemble_one_year_input(ev_largebattery, climate)

    elif kind == "ev_commercial":
        climate = pd.read_csv("examples/climates/nsrdb_honolulu.csv")
        commercial_ev = pd.read_csv("examples/application profiles/commercial_ev.csv")
        commercial_ev = commercial_ev.iloc[np.linspace(0, 24 * 3600 * 7 - 1, 24 * 7)]
        input = assemble_one_year_input(commercial_ev, climate)

    elif kind == "ev_commercial_lowdod":
        climate = pd.read_csv("examples/climates/nsrdb_honolulu.csv")
        commercial_ev_lowdod = pd.read_csv('examples/application profiles/commercial_ev_lowdod.csv')
        commercial_ev_lowdod = commercial_ev_lowdod.iloc[np.linspace(0, 24*3600*7 - 1, 24*7)]
        input = assemble_one_year_input(commercial_ev_lowdod, climate)

    elif kind == "ev_commercial_lowdod_lowsoc":
        climate = pd.read_csv("examples/climates/nsrdb_honolulu.csv")
        commercial_ev_lowdod_lowsoc = pd.read_csv('examples/application profiles/commercial_ev_lowdod.csv')
        commercial_ev_lowdod_lowsoc = commercial_ev_lowdod_lowsoc.iloc[np.linspace(0, 24*3600*7 - 1, 24*7)]
        commercial_ev_lowdod_lowsoc['SOC'] += -0.4
        input = assemble_one_year_input(commercial_ev_lowdod_lowsoc, climate)

    else:
        raise ValueError(
            f"Expected `kind` to be one of {_allowed_kinds}. Received {kind}."
        )

    return input



# Paul Gasper, NREL
# Large format prismatic LFP-Gr cell from a large manufacturer, with >250 Ah capacity and an
# energy-to-power ratio of 6 h**-1 (high energy density cell, low power).
# Experimental test data is reported in https://doi.org/10.1016/j.est.2023.109042.

import numpy as np
from .degradation_model import BatteryDegradationModel


class Lfp_Gr_250AhPrismatic(BatteryDegradationModel):
    """
    Model predicting the degradation of large-format prismatic commercial LFP-Gr
    cells from a large manufacturer, with >250 Ah capacity and an
    energy-to-power ratio of 6 h**-1 (high energy density cell, low power).
    Experimental test data is reported in https://doi.org/10.1016/j.est.2023.109042.

    .. note::
        EXPERIMENTAL AGING DATA SUMMARY:
            Experimental test data is reported in https://doi.org/10.1016/j.est.2023.109042.
            Aging test matrix varied temperature and state-of-charge for calendar aging, and
            varied depth-of-discharge, average state-of-charge, and C-rates for cycle aging.
            Charging rate was limited to a maximum of 0.16 C at 10 Celsius, and 0.65 C at all
            higher temperatures, so the model is only fit on low rate charge data.

        MODEL SENSITIVITY
            The model predicts degradation rate versus time as a function of temperature and average
            state-of-charge and degradation rate versus equivalent full cycles (charge-throughput) as 
            a function of average state-of-charge during a cycle, depth-of-discharge, and average of the
            charge and discharge C-rates.
            Test data showed that the cycling degradation rate was nearly identical for all cells,
            despite varying temperature, average SOC, and dis/charge rates. This means that the cycle aging
            model learned the inverse temperature depedence of the calendar aging model to 'balance' the 
            the degradation rate at the tested cycling temperatures (10 Celsius to 45 Celsius). So, the model
            will likely make very poor extrapolations outside of this tested temperature range for cycling fade.

        MODEL LIMITATIONS
            Charging and discharging rates were very conservative, following cycling protocols as suggested by the
            cell manufacturer. Degradation at higher charging or discharging rates will not be simulated accurately.
    """

    def __init__(self, degradation_scalar: float = 1, label: str = "LFP-Gr 250Ah"):
        # States: Internal states of the battery model
        self.states = {
            'qLoss_t': np.array([0]),
            'qLoss_EFC': np.array([0]),
        }

        # Outputs: Battery properties derived from state values
        self.outputs = {
            'q': np.array([1]),
            'q_t': np.array([1]),
            'q_EFC': np.array([1]),
        }

        # Stressors: History of stressors on the battery
        self.stressors = {
            'delta_t_days': np.array([np.nan]), 
            't_days': np.array([0]),
            'delta_efc': np.array([np.nan]), 
            'efc': np.array([0]),
            'TdegK': np.array([np.nan]),
            'soc': np.array([np.nan]), 
            'Ua': np.array([np.nan]), 
            'dod': np.array([np.nan]), 
            'Crate': np.array([np.nan]),
        }

        # Rates: History of stressor-dependent degradation rates
        self.rates = {
            'kcal': np.array([np.nan]),
            'kcyc': np.array([np.nan]),
        }

        # Expermental range: details on the range of experimental conditions, i.e.,
        # the range we expect the model to be valid in
        self.experimental_range = {
            'cycling_temperature': [10, 45],
            'dod': [0.8, 1],
            'soc': [0, 1],
            'max_rate_charge': 0.65,
            'max_rate_discharge': 1,
        }

        # Degradation scalar - scales all state changes by a coefficient
        self._degradation_scalar = degradation_scalar
        # Label for plotting
        self._label = label

    # Nominal capacity
    @property
    def cap(self):
        return 250

    # Define life model parameters
    @property
    def _params_life(self):
        return {
            # Capacity fade parameters
            'p1': 8.37e+04,
            'p2': -5.21e+03,
            'p3': -3.56e+03,
            'pcal': 0.526,
            'p4': 4.38e-08,
            'p5': 1.55e-08,
            'p6': 1.68e-07,
            'p7': 2.19e+03,
            'p8': 1.55e+05,
            'pcyc': 0.828,
        }
    
    def update_rates(self, stressors):
        # Calculate and update battery degradation rates based on stressor values
        # Inputs:
        #   stressors (dict): output from extract_stressors

        # Unpack stressors
        t_secs = stressors["t_secs"]
        delta_t_secs = t_secs[-1] - t_secs[0]
        TdegK = stressors["TdegK"]
        soc = stressors["soc"]
        Ua = stressors["Ua"]
        dod = stressors["dod"]
        Crate = stressors["Crate"]
        
        # Grab parameters
        p = self._params_life

        # Calculate the degradation coefficients
        kcal = (np.abs(p['p1'])
            * np.exp(p['p2']/TdegK)
            * np.exp(p['p3']*Ua/TdegK)
        )
        kcyc = ((p['p4'] + p['p5']*dod + p['p6']*Crate)
              * (np.exp(p['p7']/TdegK) + np.exp(-p['p8']/TdegK)))
        
        # Calculate time based average of each rate
        kcal = np.trapz(kcal, x=t_secs) / delta_t_secs
        kcyc = np.trapz(kcyc, x=t_secs) / delta_t_secs

        # Store rates
        rates = np.array([kcal, kcyc])
        for k, v in zip(self.rates.keys(), rates):
            self.rates[k] = np.append(self.rates[k], v)
    
    def update_states(self, stressors):
        # Update the battery states, based both on the degradation state as well as the battery performance
        # at the ambient temperature, T_celsius
        # Inputs:
            #   stressors (dict): output from extract_stressors
            
        # Unpack stressors
        delta_t_days = stressors["delta_t_days"]
        delta_efc = stressors["delta_efc"]
        
        # Grab parameters
        p = self._params_life

        # Grab rates, only keep most recent value
        r = self.rates.copy()
        for k, v in zip(r.keys(), r.values()):
            r[k] = v[-1]

        # Calculate incremental state changes
        states = self.states
        # Capacity
        dq_t = self._degradation_scalar * self._update_power_state(states['qLoss_t'][-1], delta_t_days, r['kcal'], p['pcal'])
        dq_EFC = self._degradation_scalar * self._update_power_state(states['qLoss_EFC'][-1], delta_efc, r['kcyc'], p['pcyc'])

        # Accumulate and store states
        dx = np.array([dq_t, dq_EFC])
        for k, v in zip(states.keys(), dx):
            x = self.states[k][-1] + v
            self.states[k] = np.append(self.states[k], x)
    
    def update_outputs(self, stressors):
        # Calculate outputs, based on current battery state
        states = self.states

        # Capacity
        q_t = 1 - states['qLoss_t'][-1]
        q_EFC = 1 - states['qLoss_EFC'][-1]
        q = 1 - states['qLoss_t'][-1] - states['qLoss_EFC'][-1]

        # Assemble output
        out = np.array([q, q_t, q_EFC])
        # Store results
        for k, v in zip(list(self.outputs.keys()), out):
            self.outputs[k] = np.append(self.outputs[k], v)



# Paul Gasper, NREL
# This model is replicated as reported by Schmalsteig et al, J. Power Sources 257 (2014) 325-334
# http://dx.doi.org/10.1016/j.jpowsour.2014.02.012

import numpy as np
from blast.models.degradation_model import BatteryDegradationModel


class Nmc111_Gr_Sanyo2Ah_Battery(BatteryDegradationModel):
    """
    Model predicting the degradation of Sanyo UR18650E cells, published by Schmalsteig et al:
    http://dx.doi.org/10.1016/j.jpowsour.2014.02.012.
    More detailed analysis of cell performance and voltage vs. state-of-charge data was copied from
    Ecker et al: http://dx.doi.org/10.1016/j.jpowsour.2013.09.143 (KNEE POINTS OBSERVED IN ECKER ET AL
    AT HIGH DEPTH OF DISCHARGE WERE SIMPLY NOT ADDRESSED DURING MODEL FITTING BY SCHMALSTEIG ET AL).
    Voltage lookup table here use data from Ecker et al for 0/10% SOC, and other values were extracted
    from Figure 1 in Schmalsteig et al using WebPlotDigitizer.

    .. note::
        EXPERIMENTAL AGING DATA SUMMARY:
            Calendar aging varied SOC at 50 Celsius, and temperature at 50% state-of-charge.
            Cycle aging varied depth-of-discharge and average state-of-charge at 35 Celsius at
            charge and discharge rates of 1C. 
            Relative discharge capacity is reported from measurements recorded at 35 Celsius and 1C rate.
            Relative DC resistance is reported after fitting of 10s 1C discharge pulses near 50% state-of-charge.

        MODEL SENSITIVITY
            The model predicts degradation rate versus time as a function of temperature and average
            state-of-charge and degradation rate versus equivalent full cycles (charge-throughput) as 
            a function of average voltage and depth-of-discharge.

        MODEL LIMITATIONS
            Cycle degradation predictions are NOT SENSITIVE TO TEMPERATURE OR C-RATE. Cycling degradation predictions
            are ONLY ACCURATE NEAR 1C RATE AND 35 CELSIUS CELL TEMPERATURE. 
    """
    
    def __init__(self, degradation_scalar: float = 1, label: str = "NMC111-Gr Sanyo"):
        # States: Internal states of the battery model
        self.states = {
            'qLoss_t': np.array([0]),
            'qLoss_EFC': np.array([0]),
            'rGain_t': np.array([0]),
            'rGain_EFC': np.array([0]),
        }

        # Outputs: Battery properties derived from state values
        self.outputs = {
            'q': np.array([1]),
            'q_t': np.array([1]),
            'q_EFC': np.array([1]),
            'r': np.array([1]),
            'r_t': np.array([1]),
            'r_EFC': np.array([1]),
        }

        # Stressors: History of stressors on the battery
        self.stressors = {
            'delta_t_days': np.array([np.nan]), 
            't_days': np.array([0]),
            'delta_efc': np.array([np.nan]), 
            'efc': np.array([0]),
            'TdegK': np.array([np.nan]),
            'soc': np.array([np.nan]),
            'Vrms': np.array([np.nan]),
            'dod': np.array([np.nan]),
        }

        # Rates: History of stressor-dependent degradation rates
        self.rates = {
            'alpha_cap': np.array([np.nan]),
            'beta_cap': np.array([np.nan]),
            'alpha_res': np.array([np.nan]),
            'beta_res': np.array([np.nan]),
        }

        # Expermental range: details on the range of experimental conditions, i.e.,
        # the range we expect the model to be valid in
        self.experimental_range = {
            'cycling_temperature': [20, 40],
            'dod': [0, 1],
            'soc': [0, 1],
            'max_rate_charge': 1,
            'max_rate_discharge': 1,
        }

        # Degradation scalar - scales all state changes by a coefficient
        self._degradation_scalar = degradation_scalar
        # Label for plotting
        self._label = label

    # Nominal capacity
    @property
    def cap(self):
        return 2.15

    # SOC index
    @property
    def _soc_index(self):
        return np.array([0,0.008637153,0.026779514,0.044921875,0.063064236,0.081206597,0.099348958,0.117491319,0.135633681,0.153776042,0.171918403,0.190060764,0.208203125,0.226345486,0.244487847,0.262630208,0.280772569,0.298914931,0.317057292,0.335199653,0.353342014,0.371484375,0.389626736,0.407769097,0.425911458,0.444053819,0.462196181,0.480338542,0.498480903,0.516623264,0.534765625,0.552907986,0.571050347,0.589192708,0.607335069,0.625477431,0.643619792,0.661762153,0.679904514,0.698046875,0.716189236,0.734331597,0.752473958,0.770616319,0.788758681,0.806901042,0.825043403,0.843185764,0.861328125,0.879470486,0.897612847,0.915755208,0.933897569,0.952039931,0.970182292,0.988324653,0.998220486,1])
    
    # OCV
    @property
    def _ocv(self):
        return np.array([3.331,3.345014187,3.37917149,3.411603677,3.440585632,3.466289865,3.490268982,3.511315401,3.529946658,3.547197821,3.561688798,3.574972194,3.586357962,3.597053683,3.605506753,3.613442288,3.620342753,3.626380661,3.632073544,3.63690387,3.642079219,3.646909545,3.652084894,3.657605266,3.663470662,3.670198615,3.677444104,3.686759732,3.696420384,3.708323686,3.720572012,3.734372943,3.749553967,3.765252525,3.781123596,3.797857224,3.814590852,3.832532062,3.849783226,3.866861877,3.884458064,3.900846669,3.917752809,3.934831461,3.953462717,3.971921462,3.991415276,4.011771649,4.03195551,4.05196686,4.070770628,4.087849279,4.104237884,4.120108955,4.135980025,4.152541142,4.160649189,4.162])

    # Voltage lookup table
    def calc_voltage(self, soc):
        # calculate cell voltage from soc vector
        return np.interp(soc, self._soc_index, self._ocv, left=self._ocv[0], right=self._ocv[-1])

    # Define life model parameters
    @property
    def _params_life(self):
        return {
            # Capacity fade parameters
            'qcal_A_V': 7.543,
            'qcal_B_V': -23.75,
            'qcal_C_TdegK': -6976,
            'qcal_p': 0.75,
            'qcyc_A_V': 7.348e-3,
            'qcyc_B_V': 3.667,
            'qcyc_C': 7.6e-4,
            'qcyc_D_DOD': 4.081e-3,
            'qcyc_p': 0.5,

            # Resistance growth parameters
            'rcal_A_V': 5.270,
            'rcal_B_V': -16.32,
            'rcal_C_TdegK': -5986,
            'rcal_p': 0.75,
            'rcyc_A_V': 2.153e-4,
            'rcyc_B_V': 3.725,
            'rcyc_C': -1.521e-5,
            'rcyc_D_DOD': 2.798e-4,
        }
        
    # Battery model
    def update_rates(self, stressors):
        # Calculate and update battery degradation rates based on stressor values
        # Inputs:
        #   stressors (dict): output from extract_stressors

        # Unpack stressors
        t_secs = stressors["t_secs"]
        delta_t_secs = t_secs[-1] - t_secs[0]
        TdegK = stressors["TdegK"]
        soc = stressors["soc"]
        Ua = stressors["Ua"]
        dod = stressors["dod"]
        Crate = stressors["Crate"]
        # Calculate RMS voltage, charge throughput
        V_rms = np.sqrt(np.mean(self.calc_voltage(soc)**2))

        # Grab parameters
        p = self._params_life

        # Calculate the degradation coefficients
        alpha_cap = (p['qcal_A_V'] * self.calc_voltage(soc) + p['qcal_B_V']) * 1e6 * np.exp(p['qcal_C_TdegK'] / TdegK)
        alpha_res = (p['rcal_A_V'] * self.calc_voltage(soc) + p['rcal_B_V']) * 1e5 * np.exp(p['rcal_C_TdegK'] / TdegK)
        beta_cap  = (
            p['qcyc_A_V'] * (V_rms - p['qcyc_B_V']) ** 2
            + p['qcyc_C']
            + p['qcyc_D_DOD'] * dod
        )
        beta_res  = (
            p['rcyc_A_V'] * (V_rms - p['rcyc_B_V']) ** 2
            + p['rcyc_C']
            + p['rcyc_D_DOD'] * dod
        )

        # Calculate time based average of each rate
        alpha_cap = np.trapz(alpha_cap, x=t_secs) / delta_t_secs
        alpha_res = np.trapz(alpha_res, x=t_secs) / delta_t_secs

        # Store rates
        rates = np.array([alpha_cap, beta_cap, alpha_res, beta_res])
        for k, v in zip(self.rates.keys(), rates):
            self.rates[k] = np.append(self.rates[k], v)
    
    def update_states(self, stressors):
        # Update the battery states, based both on the degradation state as well as the battery performance
        # at the ambient temperature, T_celsius
        # Inputs:
            #   stressors (dict): output from extract_stressors
            
        # Unpack stressors
        delta_t_days = stressors["delta_t_days"]
        delta_efc = stressors["delta_efc"]
        Ah_throughput = delta_efc * 2 * self.cap

        # Grab parameters
        p = self._params_life

        # Grab rates, only keep most recent value
        r = self.rates.copy()
        for k, v in zip(r.keys(), r.values()):
            r[k] = v[-1]

        # Calculate incremental state changes
        states = self.states
        # Capacity
        dq_t = self._degradation_scalar * self._update_power_state(states['qLoss_t'][-1], delta_t_days, r['alpha_cap'], p['qcal_p'])
        dq_EFC = self._degradation_scalar * self._update_power_state(states['qLoss_EFC'][-1], Ah_throughput, r['beta_cap'], p['qcyc_p'])
        # Resistance
        dr_t = self._degradation_scalar * self._update_power_state(states['rGain_t'][-1], delta_t_days, r['alpha_res'], p['rcal_p'])
        dr_EFC = self._degradation_scalar * r['beta_res'] * Ah_throughput

        # Accumulate and store states
        dx = np.array([dq_t, dq_EFC, dr_t, dr_EFC])
        for k, v in zip(states.keys(), dx):
            x = self.states[k][-1] + v
            self.states[k] = np.append(self.states[k], x)
    
    def update_outputs(self, stressors):
        # Calculate outputs, based on current battery state
        states = self.states
        p = self._params_life

        # Capacity
        q_t = 1 - states['qLoss_t'][-1]
        q_EFC = 1 - states['qLoss_EFC'][-1]
        q = 1 - states['qLoss_t'][-1] - states['qLoss_EFC'][-1]
        
        # Resistance
        r_t = 1 + states['rGain_t'][-1]
        r_EFC = 1 + states['rGain_EFC'][-1]
        r = 1 + states['rGain_t'][-1] + states['rGain_EFC'][-1]

        # Assemble output
        out = np.array([q, q_t, q_EFC, r, r_t, r_EFC])
        # Store results
        for k, v in zip(list(self.outputs.keys()), out):
            self.outputs[k] = np.append(self.outputs[k], v)





import numpy as np
from blast.utils.rainflow import count_cycles, extract_cycles
import pandas as pd
from typing import Union


class BatteryDegradationModel:
    def __init__(self):
        self.states = {}
        """dict: Internal states of the battery model"""
        self.outputs = {}
        """dict: Battery properties derived from state values"""
        self.stressors = {
            "delta_t_days": np.array([np.nan]),
            "t_days": np.array([0]),
            "delta_efc": np.array([np.nan]),
            "efc": np.array([0]),
            "TdegK": np.array([np.nan]),
            "soc": np.array([np.nan]),
            "Ua": np.array([np.nan]),
            "dod": np.array([np.nan]),
            "Crate": np.array([np.nan]),
        }
        """dict: History of stressors on the battery"""
        self.rates = {}
        """dict: History of stressor-dependent degradation rates"""
        self.experimental_range = {}
        """dict: details on the range of experimental conditions, i.e.,
        the range we expect the model to be valid in"""

        self._label = ""
        """str: Model name for plotting"""

    # Functions for updating time-varying states
    @staticmethod
    def _update_power_state(y0: float, dx: float, k: float, p: float):
        """
        Update time-varying power state

        Args:
            TODO Paul

        Return:
            TODO Paul
        """
        if y0 == 0:
            if dx == 0:
                dydx = 0
            else:
                y0 = k * (dx**p)
                dydx = y0 / dx
        else:
            if dx == 0:
                dydx = 0
            else:
                dydx = k * p * ((y0 / k) ** ((p - 1) / p))
        return dydx * dx

    @staticmethod
    def _update_power_B_state(y0: float, dx: float, k: float, p: float):
        """
        Update time-varying power B state

        Args:
            TODO Paul

        Return:
            TODO Paul
        """
        if y0 == 0:
            if dx == 0:
                dydx = 0
            else:
                y0 = (k * dx) ** p
                dydx = y0 / dx
        else:
            if dx == 0:
                dydx = 0
            else:
                z = (y0 ** (1 / p)) / k
                dydx = (p * (k * z) ** p) / z
        return dydx * dx

    @staticmethod
    def _update_sigmoid_state(y0: float, dx: float, y_inf: float, k: float, p: float):
        """
        Update time-varying sigmoid state

        Args:
            TODO Paul

        Return:
            TODO Paul
        """
        if y0 == 0:
            if dx == 0:
                dydx = 0
            else:
                dy = 2 * y_inf * (1 / 2 - 1 / (1 + np.exp((k * dx) ** p)))
                dydx = dy / dx
        else:
            if dx == 0:
                dydx = 0
            else:
                x_inv = (1 / k) * ((np.log(-(2 * y_inf / (y0 - y_inf)) - 1)) ** (1 / p))
                z = (k * x_inv) ** p
                dydx = (2 * y_inf * p * np.exp(z) * z) / (x_inv * (np.exp(z) + 1) ** 2)
        return dydx * dx

    # Functions to calculate voltage, half-cell potentials, or other cell specific
    # features that will be needed for calculating stressors. These equations for
    # calculating anode-to-lithium potential for a graphite electrode are a common
    # example, and are taken from M. Safari and C. Delacourt, Journal of the
    # Electrochemical Society, 158(5), A562 (2011).
    @staticmethod
    def _get_Ua(soc: np.ndarray):
        """
        Calculate Ua from SOC via lithiation fraction.

        Args:
            soc (np.ndarray)    SOC profile

        Return:
            Ua
        """
        # Calculate lithiation fraction from soc
        def get_Xa(soc):
            return 8.5 * 10**-3 + soc * (0.78 - 8.5 * 10**-3)

        Xa = get_Xa(soc)
        # Calculate Ua from lithiation fraction
        return (
            0.6379
            + 0.5416 * np.exp(-305.5309 * Xa)
            + 0.044 * np.tanh(-1 * (Xa - 0.1958) / 0.1088)
            - 0.1978 * np.tanh((Xa - 1.0571) / 0.0854)
            - 0.6875 * np.tanh((Xa + 0.0117) / 0.0529)
            - 0.0175 * np.tanh((Xa - 0.5692) / 0.0875)
        )

    def simulate_battery_life(
        self,
        input_timeseries: Union[dict, pd.DataFrame],
        simulation_years: float = None,
        is_constant_input: bool = False,
        breakpoints_max_time_diff_s: float = 86400,
        breakpoints_max_EFC_diff: float = 1,
    ):
        """
        Run battery life simulation over the input, or repeat for the number of years specified.

        Updates attributes self.rates, self.stressors, self.outputs, and self.states inplace.

        Args:
            input_timeseries (dict, pd.DataFrame)   Input time, SOC, and temperature series
            simulation_years (float)                Number of years to simulate
            is_constant_input (bool)                Whether the input is constant
            breakpoints_max_time_diff_s (float)     Max time difference for finding breakpoints
            breakpoints_max_EFC_diff (float)        Max EFC difference for finding breakpoints
        """

        # Check if required keys are in the input_timeseries dictionary
        required_keys = ["Temperature_C", "SOC", "Time_s"]
        for key in required_keys:
            if key not in input_timeseries:
                raise ValueError(
                    f"Required key '{key}' is missing in the input_timeseries dictionary."
                )
        # Check if all values are numpy arrays and of the same length
        lengths = [len(input_timeseries[key]) for key in required_keys]
        if len(set(lengths)) != 1:
            raise ValueError(
                "All numpy arrays in input_timeseries must have the same length."
            )

        if is_constant_input:
            # Run life sim repeating until simulation is longer than simulation_years
            # Only calculate stressors / degradation rates on the first timestep to dramatically accelerate the simulation
            self.update_battery_state(
                input_timeseries["Time_s"],
                input_timeseries["SOC"],
                input_timeseries["Temperature_C"],
            )
            years_simulated = self.stressors["t_days"][-1] / 365
            while years_simulated < simulation_years:
                self.update_battery_state_repeating()
                years_simulated += self.stressors["t_days"][-1] / 365
            return self
        else:
            # Unpack the inputs, calculate equivalent full cycles (EFCs)
            if isinstance(input_timeseries, dict):
                t_secs = input_timeseries["Time_s"]
                soc = input_timeseries["SOC"]
                temperature = input_timeseries["Temperature_C"]
            elif isinstance(input_timeseries, pd.DataFrame):
                t_secs = input_timeseries["Time_s"].values
                soc = input_timeseries["SOC"].values
                temperature = input_timeseries["Temperature_C"].values
            else:
                raise TypeError("'input_timeseries' was not a dict or a Dataframe.")

            # Check if we need to tile the inputs, do it if we do
            if (
                simulation_years is not None
                and t_secs[-1] < simulation_years * 365 * 24 * 3600
            ):
                # Check that inputs are periodic before tiling (SOC and temperatures don't change absurdly)
                if soc[-1] - soc[0] > 0.2:
                    raise UserWarning(
                        f"Inputs are being tiled to simulate {simulation_years} years, big jumps between repeats may lead to unrealistic simulations. Consider using the 'make_inputs_periodic' helper function."
                    )
                if temperature[-1] - temperature[0] > 10:
                    raise UserWarning(
                        f"Inputs are being tiled to simulate {simulation_years} years, big jumps between repeats may lead to unrealistic simulations. Consider using the 'make_inputs_periodic' helper function."
                    )

                # Tile the input timeseries until it is 'simulation_years' long.
                # When tiling time, assume the timestep between repeats is the most common timestep.
                n_tile = np.ceil(
                    (simulation_years * 365 * 24 * 3600) / t_secs[-1]
                ).astype(int)
                delta_t_secs = np.ediff1d(t_secs, to_begin=0)
                delta_t_secs[0] = np.median(delta_t_secs)
                delta_t_secs = np.tile(delta_t_secs, n_tile)
                delta_t_secs[0] = 0
                t_secs = np.cumsum(delta_t_secs)
                soc = np.tile(soc, n_tile)
                temperature = np.tile(temperature, n_tile)

                # Cutoff once simulation is at simulation_years
                idx_end = np.argwhere(t_secs > simulation_years * 365 * 24 * 3600)
                idx_end = idx_end[0][0]
                t_secs = t_secs[:idx_end]
                soc = soc[:idx_end]
                temperature = temperature[:idx_end]

            # Chunk the input timeseries into cycles (SOC turning points) using the rainflow algorithm
            cycles_generator = (
                [span, mean, count, idx_start, idx_end]
                for span, mean, count, idx_start, idx_end in extract_cycles(soc)
            )
            idx_turning_point = []
            for cycle in cycles_generator:
                idx_turning_point.append(cycle[-1])
            idx_turning_point = np.array(idx_turning_point)

            # Low DOD cycles can be extremely short, or the cycle time during long periods of storage excessively long.
            # Find breakpoints for simulating degradation, either simulating degradation when the number of EFCs between
            # turning points is greater than 1, or every day if EFCs accumlate slower than that.
            EFCs = np.cumsum(np.abs(np.ediff1d(soc, to_begin=0))) / 2
            simulation_breakpoints = self._find_breakpoints(
                t_secs,
                EFCs,
                idx_turning_point,
                max_time_diff_s=breakpoints_max_time_diff_s,
                max_EFC_diff=breakpoints_max_EFC_diff,
            )
            # Add the final point manually if it's not there by coinkydink
            if simulation_breakpoints[-1] != len(t_secs) - 1:
                simulation_breakpoints.append(len(t_secs) - 1)

            # Simulate battery life between breakpoints until we reach the end
            prior_breakpoint = 0
            for breakpoint in simulation_breakpoints:
                self.update_battery_state(
                    t_secs[prior_breakpoint:breakpoint],
                    soc[prior_breakpoint:breakpoint],
                    temperature[prior_breakpoint:breakpoint],
                )
                prior_breakpoint = breakpoint

    @staticmethod
    def _find_breakpoints(
        time_seconds: np.ndarray,
        EFCs: np.ndarray,
        idx_turning_point: np.ndarray,
        max_time_diff_s=86400,
        max_EFC_diff=1,
    ):
        """
        Find breakpoints for simulating battery degradation, where degradation is calculated once
        a certain number of equivalent full cycles has passed, otherwise after a certain time has passed.
        Defaults are to calculate degradation every 1 EFC or every day.

        Args:
            time_seconds (np.ndarray)       Time in seconds
            EFCs (np.ndarray)               Equivalent full cycles
            idx_turning_point (np.ndarray)  List of indices in the SOC profile that represent
                                            the end of cycles
            max_time_diff_s (int)           Maximum time difference in seconds
            max_EFC_diff (int)              Maximum difference in equivalent full cycles

        Return:
            breakpoints (list)
        """

        breakpoints = []
        last_breakpoint = 0
        for i in idx_turning_point:
            # Check if change in time_seconds greater than max_time_diff_s
            if time_seconds[i] - time_seconds[last_breakpoint] > max_time_diff_s:
                # Add a breakpoint at each day
                for i_time in range(last_breakpoint, i):
                    if (
                        time_seconds[i_time] - time_seconds[last_breakpoint]
                        > max_time_diff_s
                    ):
                        breakpoints.append(i_time)
                        last_breakpoint = i_time
                continue

            # Check change in EFCs greater than max_EFC_diff
            if EFCs[i] - EFCs[last_breakpoint] > max_EFC_diff:
                breakpoints.append(i)
                last_breakpoint = i

        return breakpoints

    def update_battery_state(
        self, t_secs: np.ndarray, soc: np.ndarray, T_celsius: np.ndarray
    ):
        """
        Update the battery states, based both on the degradation state as well as the battery performance
        at the ambient temperature, T_celsius. This function assumes battery load is changing all the time.

        Args:
            t_secs (np.ndarray)     Vector of the time in seconds since beginning of life
                                    for the soc_timeseries data points
            soc (np.ndarray):       Vector of the state-of-charge of the battery at each t_sec
            T_celsius (ndarray)     Temperature of the battery during this time period, in Celsius units.

        """

        # Check some input types:
        if not isinstance(t_secs, np.ndarray):
            raise TypeError('Input "t_secs" must be a numpy.ndarray')
        if not isinstance(soc, np.ndarray):
            raise TypeError('Input "soc" must be a numpy.ndarray')
        if not isinstance(T_celsius, np.ndarray):
            raise TypeError('Input "T_celsius" must be a numpy.ndarray')
        if not (len(t_secs) == len(soc) and len(t_secs) == len(T_celsius)):
            raise ValueError("All input timeseries must be the same length")

        stressors = self._extract_stressors(t_secs, soc, T_celsius)
        # Unpack and store some stressors for debugging or plotting
        delta_t_days = stressors["delta_t_days"]
        delta_efc = stressors["delta_efc"]
        TdegK = stressors["TdegK"]
        soc = stressors["soc"]
        Ua = stressors["Ua"]
        dod = stressors["dod"]
        Crate = stressors["Crate"]
        t_days = self.stressors["t_days"][-1] + delta_t_days
        efc = self.stressors["efc"][-1] + delta_efc
        stressors_norm = np.array(
            [
                delta_t_days,
                t_days,
                delta_efc,
                efc,
                np.mean(TdegK),
                np.mean(soc),
                np.mean(Ua),
                dod,
                Crate,
            ]
        )
        for k, v in zip(self.stressors.keys(), stressors_norm):
            self.stressors[k] = np.append(self.stressors[k], v)

        self.update_rates(stressors)
        self.update_states(stressors)
        self.update_outputs(stressors)

    def update_battery_state_repeating(self):
        """
        Update the battery states, based both on the degradation state as well as the battery performance
        at the ambient temperature, T_celsius. This function assumes battery load is repeating, i.e., stressors and
        degradation rates are unchanging for every timestep, and don't need to be calculated again.

        Updates self.states and self.outputs inplace.
        """

        # Just accumulate the cumulative stressors (total time, charge throughput in units of EFCs)
        self.stressors["t_days"] = np.append(
            self.stressors["t_days"],
            self.stressors["t_days"][-1] + self.stressors["delta_t_days"][-1],
        )
        self.stressors["efc"] = np.append(
            self.stressors["efc"],
            self.stressors["efc"][-1] + self.stressors["delta_efc"][-1],
        )

        # copy end values of old stressors into a dict
        stressors = {}
        for k, v in zip(self.stressors.keys(), self.stressors.values()):
            stressors[k] = v[-1]

        self.update_states(stressors)
        self.update_outputs(stressors)

    def update_rates(self, stressors: dict):
        """
        Calculate and update battery degradation rates based on stressor values

        Updates self.rates inplace.

        Args:
            stressors (dict)    Output from extract_stressors()
        """
        pass

    def update_states(self, stressors: dict):
        """
        Update the battery states, based both on the degradation state as well as the battery performance
        at the ambient temperature, T_celsius

        Updates self.states inplace.

        Args:
            stressors (dict)    Output from extract_stressors()
        """
        pass

    def update_outputs(self, stressors: dict):
        """
        Calculate outputs, based on current battery state (and maybe stressors)

        Updates self.outputs inplace.

        Args:
            stressors (dict)    Output from extract_stressors()
        """
        pass

    @staticmethod
    def _extract_stressors(
        t_secs: np.ndarray, soc: np.ndarray, T_celsius: np.ndarray
    ) -> dict:
        """Extract stressor values including time, temperature, depth-of-discharge,
        temperature, Ua, C-rate, and cycles.

        Args:
            t_secs (np.ndarray)     Array of time in seconds
            soc (np.ndarray)        Array of SOC
            T_celsius (np.ndarray)  Array of temperatures in C

        Return:
            stressors (dict)        Dictionary of stressor values.

        """
        # Extract stressors
        t_days = t_secs / (24 * 60 * 60)
        delta_t_days = t_days[-1] - t_days[0]
        delta_efc = (
            np.sum(np.abs(np.ediff1d(soc, to_begin=0))) / 2
        )  # sum the total changes to SOC / 2
        dod = np.max(soc) - np.min(soc)
        abs_instantaneous_crate = np.abs(
            np.diff(soc) / np.diff(t_secs / (60 * 60))
        )  # get instantaneous C-rates
        abs_instantaneous_crate[
            abs_instantaneous_crate < 1e-2
        ] = 0  # get rid of extremely small values (storage) before calculating mean
        Crate = np.trapz(abs_instantaneous_crate, t_days[1:]) / delta_t_days
        # Check storage condition, which will give nan Crate:
        if np.isnan(Crate):
            Crate = 0
        T_kelvin = T_celsius + 273.15
        # Estimate Ua (anode to reference potential) from SOC.
        # Uses the equation from Safari and Delacourt, https://doi.org/10.1149/1.3567007.
        # Anode stoichiometry is assumed to be the same for any chemistry/cell, and is calculated using the equation from Schimpe et al https://doi.org/10.1149/2.1181714jes
        # While this will not be precise, it still helps get a guess as to where the plateaus of the anode-reference potential are.
        Ua = BatteryDegradationModel._get_Ua(soc)

        cycles = count_cycles(soc)
        cycles = sum(i for _, i in cycles)

        stressors = {
            "t_secs": t_secs,
            "delta_t_days": delta_t_days,
            "delta_efc": delta_efc,
            "TdegK": T_kelvin,
            "soc": soc,
            "dod": dod,
            "Ua": Ua,
            "Crate": Crate,
            "cycles": cycles,
        }

        return stressors


"""
# Test of 'find_breakpoints'
from .rainflow import extract_cycles

hours = 6 * 24
t_hours = np.arange(hours + 1)
t_secs = t_hours * 3600
soc = np.tile(
    np.array(
        [1, 0.6, 0.4, 0.4, 0.6, 1,
        1, 0.5, 0, 0.5, 1, 1,
        1, 1, 1, 1, 1, 1,
        1, 1, 1, 1, 1, 1,
        0.98, 1, 0.98, 1, 0.98, 1,
        0.98, 1, 0.98, 1, 0.98, 1,
        0.98, 1, 0.98, 1, 0.98, 1,
        0.98, 1, 0.98, 1, 0.98, 1,
        0.98, 1, 0.98, 1, 0.98, 1,
        0.98, 1, 0.98, 1, 0.98, 1,
        0.98, 1, 0.98, 1, 0.98, 1,
        0.98, 1, 0.98, 1, 0.98, 1,]
        ),
    2
)
soc = np.append(soc, 1)
EFCs = np.cumsum(np.abs(np.ediff1d(soc, to_begin=0)))/2

# Chunk the input timeseries into cycles (SOC turning points) using the rainflow algorithm
cycles_generator = (
        [span, mean, count, idx_start, idx_end] for span, mean, count, idx_start, idx_end in extract_cycles(soc)
    )
idx_turning_point = []
for cycle in cycles_generator:
    idx_turning_point.append(cycle[-1])
idx_turning_point = np.array(idx_turning_point)

# Find the breakpoints for battery life simulation
idx_simulation_breakpoints = find_breakpoints(t_secs, EFCs, idx_turning_point)

plt.plot(t_hours/24, soc)
plt.plot(t_hours[idx_turning_point]/24, soc[idx_turning_point], '.g')
plt.plot(t_hours[idx_simulation_breakpoints]/24, soc[idx_simulation_breakpoints], 'or')
"""



# Paul Gasper, NREL
# Model predicting the degradation of Sony-Murata US18650VTC5A 3.5 Ah NCA-GrSi cylindrical cells.
# Relatively high power cells, with 1.4 wt% Si in the graphite-si composite negative electrode.
# (Maximum continuous charge rate of 2C and max continuous discharge rate of 10C, even at 5 degC)
# Data is from Technical University of Munich, reported in studies led by Leo Wildfeuer and Alexander Karger.
# Accelerated aging test data reported in https://doi.org/10.1016/j.jpowsour.2022.232498
# The model here was identifed using AI-Batt at NREL. Note that the TUM authors have put more effort
# into model identification on this data set, see the following papers for more detailed models identified on this data set:
#   Mechanistic cycle aging model (LLI, LAM_PE, LAM_(NE,Gr), LAM_(NE,Si)): https://doi.org/10.1016/j.jpowsour.2023.233947
#   Mechanistic calendar aging model (considers impact of capacity check frequency): https://doi.org/10.1016/j.jpowsour.2023.233208

import numpy as np
from blast.models.degradation_model import BatteryDegradationModel


class NCA_GrSi_SonyMurata2p5Ah_Battery(BatteryDegradationModel):

    """
    Model predicting the degradation of Sony-Murata US18650VTC5A 3.5 Ah NCA-GrSi cylindrical cells.
    Relatively high power cells, with 1.4 wt% Si in the graphite-si composite negative electrode.
    (Maximum continuous charge rate of 2C and max continuous discharge rate of 10C, even at 5 degC)
    Data is from Technical University of Munich, reported in studies led by Leo Wildfeuer and Alexander Karger.
    Accelerated aging test data reported in https://doi.org/10.1016/j.jpowsour.2022.232498
    The model here was identifed using AI-Batt at NREL. Note that the TUM authors have put more effort
    into model identification on this data set, see the following papers for more detailed models identified on this data set:
    Mechanistic cycle aging model (LLI, LAM_PE, LAM_(NE,Gr), LAM_(NE,Si)): https://doi.org/10.1016/j.jpowsour.2023.233947
    Mechanistic calendar aging model (considers impact of capacity check frequency): https://doi.org/10.1016/j.jpowsour.2023.233208
    
    .. note::
        EXPERIMENTAL AGING DATA SUMMARY:
            Aging test matrix varied temperature and state-of-charge for calendar aging, and
            varied depth-of-discharge, average state-of-charge, and C-rates for cycle aging.
            For calendar aging, in addition to temperature and SOC, capacity-check frequency was also varied. 

        MODEL SENSITIVITY
            The model predicts degradation rate versus time as a function of temperature and average
            state-of-charge and degradation rate versus equivalent full cycles (charge-throughput) as 
            a function of average state-of-charge during a cycle, depth-of-discharge, and average of the
            charge and discharge C-rates.

        MODEL LIMITATIONS
            I did not model the impact of capacity check frequency, only using the 6 week capacity check data.
            See https://doi.org/10.1016/j.jpowsour.2023.233208 for a detailed consideration of the capacity check frequency impact on aging.
            Only one cycling cell in the data shows 'knee-over' behavior, making empirical model identification of this behavior challenging.
            I simply neglected the knee over; this knee occurs at ~80% capacity fade, so note that predictions of degradation at maximum
            DOD and below 80% capacity should not be believed.
    """

    def __init__(self, degradation_scalar: float = 1, label: str = "NCA-GrSi Sony-Murata"):
        # States: Internal states of the battery model
        self.states = {
            'qLoss_t': np.array([0]),
            'qLoss_EFC': np.array([0]),
        }

        # Outputs: Battery properties derived from state values
        self.outputs = {
            'q': np.array([1]),
            'q_t': np.array([1]),
            'q_EFC': np.array([1]),
        }

        # Stressors: History of stressors on the battery
        self.stressors = {
            'delta_t_days': np.array([np.nan]), 
            't_days': np.array([0]),
            'delta_efc': np.array([np.nan]), 
            'efc': np.array([0]),
            'TdegK': np.array([np.nan]),
            'soc': np.array([np.nan]), 
            'soc_low': np.array([np.nan]),
            'soc_high': np.array([np.nan]),
            'Ua': np.array([np.nan]), 
            'dod': np.array([np.nan]), 
            'Cdis': np.array([np.nan]),
            'Cchg': np.array([np.nan]),
        }

        # Rates: History of stressor-dependent degradation rates
        self.rates = {
            'kcal': np.array([np.nan]),
            'pcal': np.array([np.nan]),
            'kcyc': np.array([np.nan]),
            'pcyc': np.array([np.nan]),
        }

        # Expermental range: details on the range of experimental conditions, i.e.,
        # the range we expect the model to be valid in
        self.experimental_range = {
            'cycling_temperature': [5, 35],
            'dod': [0.2, 1],
            'soc': [0, 1],
            'max_rate_charge': 2,
            'max_rate_discharge': 10,
        }

        # Degradation scalar - scales all state changes by a coefficient
        self._degradation_scalar = degradation_scalar
        # Label for plotting
        self._label = label

    # Nominal capacity
    @property
    def cap(self):
        return 2.5

    # Define life model parameters
    @property
    def _params_life(self):
        return {
            # Capacity fade parameters
            'q2': 3.41,
            'q1_b0': 0.00748,
            'q1_b1': 8.9,
            'q1_b2': -4.87,
            'p_cal_b0': 1.01,
            'p_cal_b1': -0.133,
            'p_cal_b2': 1.22,
            'p_cal_b3': -0.56,
            'q4': 0.479,
            'q3_b0': 5.49,
            'q3_b1': -64.9,
            'q3_b2': 21.3,
            'q3_b3': 0.000801,
            'p_cyc_b0': 0.422,
            'p_cyc_b1': 2.64,
            'p_cyc_b2': 0.00627,
            'p_cyc_b3': -0.0342,
            'p_cyc_b4': 1.29,
            'p_cyc_b5': -2.14
        }

    # Override super class update_battery_state because this model uses a few unusual stressors
    def update_battery_state(self, t_secs, soc, T_celsius):
        # Update the battery states, based both on the degradation state as well as the battery performance
        # at the ambient temperature, T_celsius. This function assumes battery load is changing all the time.
        # Inputs:
            #   t_secs (ndarry): vector of the time in seconds since beginning of life for the soc_timeseries data points
            #   soc (ndarry): vector of the state-of-charge of the battery at each t_sec
            #   T_celsius (ndarray): the temperature of the battery during this time period, in Celsius units.
            
        # Check some input types:
        if not isinstance(t_secs, np.ndarray):
            raise TypeError('Input "t_secs" must be a numpy.ndarray')
        if not isinstance(soc, np.ndarray):
            raise TypeError('Input "soc" must be a numpy.ndarray')
        if not isinstance(T_celsius, np.ndarray):
            raise TypeError('Input "T_celsius" must be a numpy.ndarray')
        if not (len(t_secs) == len(soc) and len(t_secs) == len(T_celsius)):
            raise ValueError('All input timeseries must be the same length')
        
        stressors = self._extract_stressors(t_secs, soc, T_celsius)
        # Unpack and store some stressors for debugging or plotting
        delta_t_days = stressors["delta_t_days"]
        delta_efc = stressors["delta_efc"]
        TdegK = stressors["TdegK"]
        soc = stressors["soc"]
        Ua = stressors["Ua"]
        dod = stressors["dod"]
        soc_low = stressors["soc_low"]
        soc_high = stressors["soc_high"]
        Cchg = stressors["Cchg"]
        Cdis = stressors["Cdis"]
        t_days = self.stressors['t_days'][-1] + delta_t_days
        efc = self.stressors['efc'][-1] + delta_efc
        stressors_norm = np.array([delta_t_days, t_days, delta_efc, efc, np.mean(TdegK), np.mean(soc), soc_low, soc_high, np.mean(Ua), dod, Cdis, Cchg])
        for k, v in zip(self.stressors.keys(), stressors_norm):
            self.stressors[k] = np.append(self.stressors[k], v)
            
        self.update_rates(stressors)
        # print(f"kcal: {self.rates['kcal'][-1]}, pcal: {self.rates['pcal'][-1]}, kcyc: {self.rates['kcyc'][-1]}, pcyc: {self.rates['pcyc'][-1]}")
        self.update_states(stressors)
        # print(f"qLoss_t: {self.states['qLoss_t'][-1]}, qLoss_EFC: {self.states['qLoss_EFC'][-1]}")
        self.update_outputs(stressors)
        # print(f"q_t: {self.outputs['q_t'][-1]}, q_EFC: {self.outputs['q_EFC'][-1]}, q: {self.outputs['q'][-1]}")

    def update_rates(self, stressors):
        # Calculate and update battery degradation rates based on stressor values
        # Inputs:
        #   stressors (dict): output from extract_stressors

        # Unpack stressors
        t_secs = stressors["t_secs"]
        delta_t_secs = t_secs[-1] - t_secs[0]
        TdegK = stressors["TdegK"]
        soc = stressors["soc"]
        soc_low = stressors["soc_low"]
        soc_high = stressors["soc_high"]
        Ua = stressors["Ua"]
        dod = stressors["dod"]
        Cdis = stressors["Cdis"]
        Cchg = stressors['Cchg']
        # Normalized temperature and Ua
        TdegKN = TdegK / (273.15+45)
        UaN = Ua / 0.123 # Ua at about 50% SOC for Gr

        # Grab parameters
        p = self._params_life

        # Calculate the degradation coefficients
        kcal = (p['q1_b0']
                * np.exp(p['q1_b1']*(TdegKN**2)*(1/(UaN**(1/3))))
                * np.exp(p['q1_b2']*(TdegKN**2)*(1/(UaN**0.5)))
            )
        pcal = (p['p_cal_b0']
            * np.exp(p['p_cal_b1']*(TdegKN**3))
            * np.exp(p['p_cal_b2']*(TdegKN**3)*(soc**3))
            * np.exp(p['p_cal_b3']*(TdegKN**2)*(1/(UaN**3)))
            )
        kcyc = (p['q3_b0']
                + p['q3_b1']*(Cchg**0.5)*(1/(TdegKN**3))*soc
                + p['q3_b2']*np.exp((Cchg**0.5)*(1/(TdegKN**3))*soc)
                + p['q3_b3']*np.exp(Cdis*(1/(TdegKN**3))*(soc**0.5))
            )
        pcyc = np.abs(p['p_cyc_b0']
                    + p['p_cyc_b1']*np.exp(Cdis*(1/(TdegKN**3))*(soc_high**0.5))
                    + p['p_cyc_b2']*np.exp((dod**0.5)*(Cchg**0.5)*(TdegKN**3)*(soc_low**2))
                    + p['p_cyc_b3']*(TdegKN**2)*soc_low
                    + p['p_cyc_b4']*(dod**0.5)*(Cdis**0.5)*(1/TdegKN)*(Cchg**0.5)
                    + p['p_cyc_b5']*np.exp((dod**0.5)*(Cchg**3)*TdegKN*(Cdis**0.5))
            )
        
        # Calculate time based average of each rate
        kcal = np.trapz(kcal, x=t_secs) / delta_t_secs
        pcal = np.trapz(pcal, x=t_secs) / delta_t_secs
        kcyc = np.trapz(kcyc, x=t_secs) / delta_t_secs
        pcyc = np.trapz(pcyc, x=t_secs) / delta_t_secs

        # kcyc linear model went negative in a few extrapolations on model investigation
        # floor at near the minimum value observed in the experimental data set
        kcyc = np.max(np.array((kcyc, 0.15)))
        if kcyc == 1.5:
            print('kcyc hit floor')

        # Store rates
        rates = np.array([kcal, pcal, kcyc, pcyc])
        for k, v in zip(self.rates.keys(), rates):
            self.rates[k] = np.append(self.rates[k], v)
    
    def update_states(self, stressors):
        # Update the battery states, based both on the degradation state as well as the battery performance
        # at the ambient temperature, T_celsius
        # Inputs:
            #   stressors (dict): output from extract_stressors
            
        # Unpack stressors
        delta_t_days = stressors["delta_t_days"]
        delta_efc = stressors["delta_efc"]
        
        # Grab parameters
        p = self._params_life

        # Grab rates, only keep most recent value
        r = self.rates.copy()
        for k, v in zip(r.keys(), r.values()):
            r[k] = v[-1]

        # Calculate incremental state changes
        states = self.states

        # Capacity
        dq_t = self._degradation_scalar * self._update_sigmoid_state(states['qLoss_t'][-1], delta_t_days/5e3, r['kcal'], p['q2'], r['pcal'])
        dq_EFC = self._degradation_scalar * self._update_sigmoid_state(states['qLoss_EFC'][-1], delta_efc/1e5, r['kcyc'], p['q4'], r['pcyc'])

        # Accumulate and store states
        dx = np.array([dq_t, dq_EFC])
        for k, v in zip(states.keys(), dx):
            x = self.states[k][-1] + v
            self.states[k] = np.append(self.states[k], x)
    
    def update_outputs(self, stressors):
        # Calculate outputs, based on current battery state
        states = self.states

        # Capacity
        q_t = 1 - states['qLoss_t'][-1]
        q_EFC = 1 - states['qLoss_EFC'][-1]
        q = 1 - states['qLoss_t'][-1] - states['qLoss_EFC'][-1]

        # Assemble output
        out = np.array([q, q_t, q_EFC])
        # Store results
        for k, v in zip(list(self.outputs.keys()), out):
            self.outputs[k] = np.append(self.outputs[k], v)
    
    @staticmethod
    def _extract_stressors(t_secs, soc, T_celsius):
        # extract the usual stressors
        stressors = BatteryDegradationModel._extract_stressors(t_secs, soc, T_celsius)
        # model specific stressors: soc_low, soc_high, Cchg, Cdis
        soc_low = np.min(soc)
        soc_high = np.max(soc)
        t_days = t_secs / (24*60*60)
        delta_t_days = t_days[-1] - t_days[0]
        dt = np.diff(t_secs)
        abs_instantaneous_crate = np.abs(np.diff(soc)/np.diff(t_secs/(60*60))) # get instantaneous C-rates
        abs_instantaneous_crate[abs_instantaneous_crate < 1e-2] = 0 # get rid of extremely small values (storage) before calculating mean
        Crate = np.trapz(abs_instantaneous_crate, t_days[1:]) / delta_t_days
        # Check storage condition, which will give nan Crate:
        if np.isnan(Crate):
            Cdis = 0
            Cchg = 0
        else:
            instantaneous_crate = np.diff(soc)/(dt/3600)
            instantaneous_crate[np.abs(instantaneous_crate) < 1e-2] = 0 # threshold tiny values to zero to prevent weirdness
            mask_cchg = instantaneous_crate > 0
            mask_cdis = instantaneous_crate < 0
            instantaneous_cchg = instantaneous_crate[mask_cchg]
            dt_charging = dt[mask_cchg]
            t_secs_charging = np.cumsum(dt_charging)
            if len(instantaneous_cchg) > 1:
                Cchg = np.trapz(instantaneous_cchg, t_secs_charging) / (t_secs_charging[-1] - t_secs_charging[0])
            elif len(instantaneous_cchg) == 1:
                Cchg = instantaneous_cchg[0]
            else: # half cycle with no charge segment
                Cchg = 0
            instantaneous_cdis = instantaneous_crate[mask_cdis]
            dt_discharging = dt[mask_cdis]
            t_secs_discharging = np.cumsum(dt_discharging)
            if len(instantaneous_cdis) > 1:
                Cdis = np.trapz(np.abs(instantaneous_cdis), t_secs_discharging) / (t_secs_discharging[-1] - t_secs_discharging[0])
            elif len(instantaneous_cdis) == 1:
                Cdis = np.abs(instantaneous_cdis)
                Cdis = Cdis[0]
            else: # half cycle with no discharge segment
                Cdis = 0
        
        stressors['soc_low'] = soc_low
        stressors['soc_high'] = soc_high
        stressors['Cchg'] = Cchg
        stressors['Cdis'] = Cdis
        return stressors




# Paul Gasper, NREL
import numpy as np
from blast.models.degradation_model import BatteryDegradationModel


class Nmc111_Gr_Kokam75Ah_Battery(BatteryDegradationModel):
    """
    Model predicting the degradation of a Kokam 75 Ah NMC-Gr pouch cell.
    https://ieeexplore.ieee.org/iel7/7951530/7962914/07963578.pdf
    It is uncertain if the exact NMC composition is 1-1-1, but it this is definitely not a high nickel (>80%) cell.
    Degradation rate is a function of the aging stressors, i.e., ambient temperature and use.
    The state of the battery is updated throughout the lifetime of the cell.
    Performance metrics are capacity and DC resistance. These metrics change as a function of the
    cell's current degradation state, as well as the ambient temperature. The model predicts time and 
    cycling dependent degradation, using Loss of Lithium Inventory (LLI) and Loss of Active
    Material (LAM) degradation modes that interact competitively (cell performance is limited by
    one or the other.)
    Parameters to modify to change fade rates:
    - Calendar capacity loss rate: q1_0
    - Cycling capacity loss rate (LLI): q3_0
    - Cycling capacity loss rate (LAM): q5_0, will also effect resistance growth onset due to LAM.
    - Calendar resistance growth rate (LLI), relative to capacity loss rate: r1
    - Cycling resistance growth rate (LLI), relative to capacity loss rate: r3

    .. note::
        EXPERIMENTAL AGING DATA SUMMARY:
            Aging test matrix varied primarly temperature, with small DOD variation.
            Calendar and cycle aging were performed between 0 and 55 Celsius. C-rates always at 1C,
            except for charging at 0 Celsius, which was conducted at C/3. Depth-of-discharge was 80%
            for nearly all tests (3.4 V - 4.1 V), with one 100% DOD test (3 V - 4.2 V).
            Reported relative capacity was measured at C/5 rate at the aging temperatures. Reported
            relative DC resistance was measured by HPPC using a 10s, 1C DC pulse, averaged between 
            charge and discharge, calculated using a simple ohmic fit of the voltage response.

        MODEL SENSITIVITY
            The model predicts degradation rate versus time as a function of temperature and average
            state-of-charge and degradation rate versus equivalent full cycles (charge-throughput) as 
            a function of temperature and depth-of-discharge. Sensitivity to cycling degradation rate
            at low temperature is inferred from physical insight due to limited data.

        MODEL LIMITATIONS
            There is NO C-RATE DEPENDENCE for degradation in this model. THIS IS NOT PHYSICALLY REALISTIC
            AND IS BASED ON LIMITED DATA.
    """

    def __init__(self, degradation_scalar: float = 1, label: str = "NMC111-Gr Kokam"):
        # States: Internal states of the battery model
        self.states = {
            'qLoss_LLI_t': np.array([0]),   # relative Li inventory change, time dependent (SEI)
            'qLoss_LLI_EFC': np.array([0]), # relative Li inventory change, charge-throughput dependent (SEI)
            'qLoss_LAM': np.array([1e-8]),  # relative active material change, charge-throughput dependent (electrode damage)
            'rGain_LLI_t': np.array([0]),   # relative SEI growth, time dependent (SEI)
            'rGain_LLI_EFC': np.array([0]), # relative SEI growth, charge-throughput dependent (SEI)
        }

        # Outputs: Battery properties derived from state values
        self.outputs = {
            'q': np.array([1]),         # relative capacity
            'q_LLI': np.array([1]),     # relative lithium inventory
            'q_LLI_t': np.array([1]),   # relative lithium inventory, time dependent loss
            'q_LLI_EFC': np.array([1]), # relative lithium inventory, charge-throughput dependent loss
            'q_LAM': np.array([1.01]),  # relative active material, charge-throughput dependent loss
            'r': np.array([1]),         # relative resistance
            'r_LLI': np.array([1]),     # relative SEI resistance
            'r_LLI_t': np.array([1]),   # relative SEI resistance, time dependent growth
            'r_LLI_EFC': np.array([1]), # relative SEI resistance, charge-throughput dependent growth
            'r_LAM': np.array([1]),     # relative electrode resistance, q_LAM dependent growth
        }

        # Stressors: History of stressors on the battery
        self.stressors = {
            'delta_t_days': np.array([np.nan]), 
            't_days': np.array([0]),
            'delta_efc': np.array([np.nan]), 
            'efc': np.array([0]),
            'TdegK': np.array([np.nan]),
            'soc': np.array([np.nan]),
            'Ua': np.array([np.nan]),
            'dod': np.array([np.nan]),
        }

        # Rates: History of stressor-dependent degradation rates
        self.rates = {
            'q1': np.array([np.nan]),
            'q3': np.array([np.nan]),
            'q5': np.array([np.nan]),
        }

        # Expermental range: details on the range of experimental conditions, i.e.,
        # the range we expect the model to be valid in
        self.experimental_range = {
            'cycling_temperature': [0, 45],
            'dod': [0.8, 1],
            'soc': [0, 1],
            'max_rate_charge': 1,
            'max_rate_discharge': 1,
        }

        # Degradation scalar - scales all state changes by a coefficient
        self._degradation_scalar = degradation_scalar
        # Label for plotting
        self._label = label
    
    # Nominal capacity
    @property
    def cap(self):
        return 75

    # Define life model parameters
    @property
    def _params_life(self):
        return {
            'q1_0'  : 2.66e7,   # CHANGE to modify calendar degradation rate (larger = faster degradation)
            'q1_1'  : -17.8,
            'q1_2'  : -5.21,
            'q2'    : 0.357,
            'q3_0'  : 3.80e3,   # CHANGE to modify cycling degradation rate (LLI) (larger = faster degradation)
            'q3_1'  : -18.4,
            'q3_2'  : 1.04,
            'q4'    : 0.778,
            'q5_0'  : 1e4,      # CHANGE to modify cycling degradation rate (LAM) (accelerating fade onset) (larger = faster degradation)
            'q5_1'  : 153,
            'p_LAM' : 10,
            'r1'    : 0.0570,   # CHANGE to modify change of resistance relative to change of capacity (calendar degradation)
            'r2'    : 1.25,
            'r3'    : 4.87,     # CHANGE to modify change of resistance relative to change of capacity (cycling degradation)
            'r4'    : 0.712,
            'r5'    : -0.08,
            'r6'    : 1.09,
        }
    
    # Battery model
    def update_rates(self, stressors):
        # Calculate and update battery degradation rates based on stressor values
        # Inputs:
        #   stressors (dict): output from extract_stressors

        # Unpack stressors
        t_secs = stressors["t_secs"]
        delta_t_secs = t_secs[-1] - t_secs[0]
        TdegK = stressors["TdegK"]
        soc = stressors["soc"]
        Ua = stressors["Ua"]
        dod = stressors["dod"]
        Crate = stressors["Crate"]
        TdegC = TdegK - 273.15
        TdegKN = TdegK / (273.15 + 35) # normalized temperature
        UaN = Ua / 0.123 # normalized anode-to-reference potential
        
        # Grab parameters
        p = self._params_life

        # Calculate degradation rates
        q1 = p['q1_0'] * np.exp(p['q1_1'] * (1 / TdegKN)) * np.exp(p['q1_2'] * (UaN / TdegKN))
        q3 = p['q3_0'] * np.exp(p['q3_1'] * (1/TdegKN)) * np.exp(p['q3_2'] * np.exp(dod**2))
        q5 = p['q5_0'] + p['q5_1'] * (TdegC - 55) * dod * (np.mean(soc)/0.6)
        # Calculate time based average of each rate
        q1 = np.trapz(q1, x=t_secs) / delta_t_secs
        q3 = np.trapz(q3, x=t_secs) / delta_t_secs
        q5 = np.trapz(q5, x=t_secs) / delta_t_secs

        # Store rates
        rates = np.array([q1, q3, q5])
        for k, v in zip(self.rates.keys(), rates):
            self.rates[k] = np.append(self.rates[k], v)

    def update_states(self, stressors):
        # Update the battery states, based both on the degradation state as well as the battery performance
        # at the ambient temperature, T_celsius
        # Inputs:
            #   stressors (dict): output from extract_stressors
            
        # Unpack stressors
        delta_t_days = stressors["delta_t_days"]
        delta_efc = stressors["delta_efc"]
        
        # Grab parameters
        p = self._params_life

        # Grab rates, only keep most recent value
        r = self.rates.copy()
        for k, v in zip(r.keys(), r.values()):
            r[k] = v[-1]
        
        # Calculate incremental state changes
        states = self.states
        # Capacity
        dq_LLI_t = self._degradation_scalar * self._update_power_state(states['qLoss_LLI_t'][-1], delta_t_days, r['q1'], p['q2'])
        dq_LLI_EFC = self._degradation_scalar * self._update_power_state(states['qLoss_LLI_EFC'][-1], delta_efc, r['q3'], p['q4'])
        dq_LAM = self._degradation_scalar * self._update_sigmoid_state(states['qLoss_LAM'][-1], delta_efc, 1, 1/r['q5'], p['p_LAM'])
        
        # Resistance
        dr_LLI_t = self._degradation_scalar * self._update_power_state(states['rGain_LLI_t'][-1], delta_t_days, p['r1']*r['q1'], p['r2'])
        dr_LLI_EFC = self._degradation_scalar * self._update_power_state(states['rGain_LLI_EFC'][-1], delta_efc, p['r3']*r['q3'], p['r4'])

        # Accumulate and store states
        dx = np.array([dq_LLI_t, dq_LLI_EFC, dq_LAM, dr_LLI_t, dr_LLI_EFC])
        for k, v in zip(states.keys(), dx):
            x = self.states[k][-1] + v
            self.states[k] = np.append(self.states[k], x)

    def update_outputs(self, stressors):
        # Calculate outputs, based on current battery state
        states = self.states
        p = self._params_life

        # Capacity
        q_LLI = 1 - states['qLoss_LLI_t'][-1] - states['qLoss_LLI_EFC'][-1]
        q_LLI_t = 1 - states['qLoss_LLI_t'][-1]
        q_LLI_EFC = 1 - states['qLoss_LLI_EFC'][-1]
        q_LAM = 1.01 - states['qLoss_LAM'][-1]
        q = np.min(np.array([q_LLI, q_LAM]))
        
        # Resistance
        r_LLI = 1 + states['rGain_LLI_t'][-1] + states['rGain_LLI_EFC'][-1]
        r_LLI_t = 1 + states['rGain_LLI_t'][-1]
        r_LLI_EFC = 1 + states['rGain_LLI_EFC'][-1]
        r_LAM = p['r5'] + p['r6'] * (1 / q_LAM)
        r = np.max(np.array([r_LLI, r_LAM]))

        # Assemble output
        out = np.array([q, q_LLI, q_LLI_t, q_LLI_EFC, q_LAM, r, r_LLI, r_LLI_t, r_LLI_EFC, r_LAM])
        # Store results
        for k, v in zip(list(self.outputs.keys()), out):
            self.outputs[k] = np.append(self.outputs[k], v)


# Paul Gasper, NREL
# Large format pouch NMC-Gr cell from a large manufacturer, with ~75 Ah capacity and an
# energy-to-power ratio of 16 h^-1 (relatively high power, suitable for fast charging).
# Experimental test data is reported in https://doi.org/10.1016/j.est.2023.109042.
# This is for the 'NMC-Gr A' cell reported in the paper.

import numpy as np
from blast.models.degradation_model import BatteryDegradationModel


class NMC_Gr_75Ah_A(BatteryDegradationModel):
    """
    Model predicting the degradation of large format pouch 'NMC-Gr A' cells from a
    large manufacturer, with ~75 Ah capacity and an energy-to-power ratio of 16 h^-1
    (relatively high power, suitable for fast charging).
    Experimental test data is reported in https://doi.org/10.1016/j.est.2023.109042.
    This is for the 'NMC-Gr A' cell reported in the paper.

    .. note::
        EXPERIMENTAL AGING DATA SUMMARY:
            Experimental test data is reported in https://doi.org/10.1016/j.est.2023.109042.
            Aging test matrix varied temperature and state-of-charge for calendar aging, and
            varied depth-of-discharge, average state-of-charge, and C-rates for cycle aging.

        MODEL SENSITIVITY
            The model predicts degradation rate versus time as a function of temperature and average
            state-of-charge and degradation rate versus equivalent full cycles (charge-throughput) as 
            a function of average state-of-charge during a cycle, depth-of-discharge, and average of the
            charge and discharge C-rates.

        MODEL LIMITATIONS
            Charging and discharging rates were conservative, following cycling protocols as suggested by the
            cell manufacturer. Degradation at higher charging or discharging rates will not be simulated accurately.
    """

    def __init__(self, degradation_scalar: float = 1, label: str = "NMC-Gr A 75Ah"):
        # States: Internal states of the battery model
        self.states = {
            'qLoss_t': np.array([0]),
            'qLoss_EFC': np.array([0]),
        }

        # Outputs: Battery properties derived from state values
        self.outputs = {
            'q': np.array([1]),
            'q_t': np.array([1]),
            'q_EFC': np.array([1]),
        }

        # Stressors: History of stressors on the battery
        self.stressors = {
            'delta_t_days': np.array([np.nan]), 
            't_days': np.array([0]),
            'delta_efc': np.array([np.nan]), 
            'efc': np.array([0]),
            'TdegK': np.array([np.nan]),
            'soc': np.array([np.nan]), 
            'Ua': np.array([np.nan]), 
            'dod': np.array([np.nan]), 
            'Crate': np.array([np.nan]),
        }

        # Rates: History of stressor-dependent degradation rates
        self.rates = {
            'kcal': np.array([np.nan]),
            'kcyc': np.array([np.nan]),
        }

        # Expermental range: details on the range of experimental conditions, i.e.,
        # the range we expect the model to be valid in
        self.experimental_range = {
            'cycling_temperature': [10, 45],
            'dod': [0.8, 1],
            'soc': [0, 1],
            'max_rate_charge': 2,
            'max_rate_discharge': 2,
        }

        # Degradation scalar - scales all state changes by a coefficient
        self._degradation_scalar = degradation_scalar
        # Label for plotting
        self._label = label

    # Nominal capacity
    @property
    def cap(self):
        return 75

    # Define life model parameters
    @property
    def _params_life(self):
        return {
            # Capacity fade parameters
            'p1': 8.71e-07,
            'p2': 10,
            'pcal': 0.578,
            'p3': 0.0267,
            'p4': 1.64,
            'p5': 0.011,
            'p6': -0.44,
            'pcyc': 1.11,
        }
    
    def update_rates(self, stressors):
        # Calculate and update battery degradation rates based on stressor values
        # Inputs:
        #   stressors (dict): output from extract_stressors

        # Unpack stressors
        t_secs = stressors["t_secs"]
        delta_t_secs = t_secs[-1] - t_secs[0]
        TdegK = stressors["TdegK"]
        soc = stressors["soc"]
        Ua = stressors["Ua"]
        dod = stressors["dod"]
        Crate = stressors["Crate"]
        TdegKN = TdegK / (293.15 + 35)
        UaN = Ua / 0.123 # Ua at about 50% SOC

        # Grab parameters
        p = self._params_life

        # Calculate the degradation coefficients
        kcal = (p['p1']
            * np.exp(p['p2']*(TdegKN**3) * (1/(UaN**(1/3)))))
        kcyc = np.abs(p['p3'] + p['p4']*(TdegKN**3)*(dod**0.5) 
        + p['p5']*np.exp((1/(TdegKN**0.5)) * (Crate**2))
        + p['p6']*np.exp((dod**0.5) * (TdegKN**2) * (Crate**0.5)))
        
        # Calculate time based average of each rate
        kcal = np.trapz(kcal, x=t_secs) / delta_t_secs
        kcyc = np.trapz(kcyc, x=t_secs) / delta_t_secs

        # Store rates
        rates = np.array([kcal, kcyc])
        for k, v in zip(self.rates.keys(), rates):
            self.rates[k] = np.append(self.rates[k], v)
    
    def update_states(self, stressors):
        # Update the battery states, based both on the degradation state as well as the battery performance
        # at the ambient temperature, T_celsius
        # Inputs:
            #   stressors (dict): output from extract_stressors
            
        # Unpack stressors
        delta_t_days = stressors["delta_t_days"]
        delta_efc = stressors["delta_efc"]
        
        # Grab parameters
        p = self._params_life

        # Grab rates, only keep most recent value
        r = self.rates.copy()
        for k, v in zip(r.keys(), r.values()):
            r[k] = v[-1]

        # Calculate incremental state changes
        states = self.states
        # Capacity
        dq_t = self._degradation_scalar * self._update_power_state(states['qLoss_t'][-1], delta_t_days/1e3, r['kcal'], p['pcal'])
        dq_EFC = self._degradation_scalar * self._update_power_state(states['qLoss_EFC'][-1], delta_efc/1e4, r['kcyc'], p['pcyc'])

        # Accumulate and store states
        dx = np.array([dq_t, dq_EFC])
        for k, v in zip(states.keys(), dx):
            x = self.states[k][-1] + v
            self.states[k] = np.append(self.states[k], x)
    
    def update_outputs(self, stressors):
        # Calculate outputs, based on current battery state
        states = self.states

        # Capacity
        q_t = 1 - states['qLoss_t'][-1]
        q_EFC = 1 - states['qLoss_EFC'][-1]
        q = 1 - states['qLoss_t'][-1] - states['qLoss_EFC'][-1]

        # Assemble output
        out = np.array([q, q_t, q_EFC])
        # Store results
        for k, v in zip(list(self.outputs.keys()), out):
            self.outputs[k] = np.append(self.outputs[k], v)



# Paul Gasper, NREL
# Model predicting the degradation of Sony-Murata 3 Ah LFP-Gr cylindrical cells.
# High power cells with low energy density and exceptional cycle life / cell quality.
# Data is from Technical University of Munich, reported in studies led by Maik Naumann.
# Capacity model identification was conducted at NREL. Resistance model is from Naumann et al.
# Naumann et al used an interative fitting procedure, but it was found that lower model error could be
# achieved by simply reoptimizing all resistance growth parameters to the entire data set.
# Calendar aging data source: https://doi.org/10.1016/j.est.2018.01.019
# Cycle aging data source: https://doi.org/10.1016/j.jpowsour.2019.227666
# Model identification source: https://doi.org/10.1149/1945-7111/ac86a8

import numpy as np
import scipy.stats as stats
from blast.models.degradation_model import BatteryDegradationModel


class Lfp_Gr_SonyMurata3Ah_Battery(BatteryDegradationModel):
    """
    Model predicting the degradation of Sony-Murata 3 Ah LFP-Gr cylindrical cells.
    Data is from Technical University of Munich, reported in studies led by Maik Naumann.
    Capacity model identification was conducted at NREL. Resistance model is from Naumann et al.
    Naumann et al used an interative fitting procedure, but it was found that lower model error could be
    achieved by simply reoptimizing all resistance growth parameters to the entire data set.
    Calendar aging data source: https://doi.org/10.1016/j.est.2018.01.019
    Cycle aging data source: https://doi.org/10.1016/j.jpowsour.2019.227666
    Model identification source: https://doi.org/10.1149/1945-7111/ac86a8

    Degradation rate is a function of the aging stressors, i.e., ambient temperature and use.
    The state of the battery is updated throughout the lifetime of the cell.
    Performance metrics are capacity and DC resistance. These metrics change as a function of the
    cell's current degradation state, as well as the ambient temperature. The model predicts time and 
    cycling dependent degradation. Cycling dependent degradation includes a break-in mechanism as well
    as long term cycling fade; the break-in mechanism strongly influenced results of the accelerated
    aging test, but is not expected to have much influence on real-world applications.
    Parameters to modify to change fade rates:
    - q1_b0: rate of capacity loss due to calendar degradation
    - q5_b0: rate of capacity loss due to cycling degradation
    - k_ref_r_cal: rate of resistance growth due to calendar degradation
    - A_r_cyc: rate of resistance growth due to cycling degradation

    .. note::
        EXPERIMENTAL AGING DATA SUMMARY:
            Aging test matrix varied temperature and state-of-charge for calendar aging, and
            varied depth-of-discharge, average state-of-charge, and C-rates for cycle aging.
            There is NO LOW TEMPERATURE cycling aging data, i.e., no lithium-plating induced by
            kinetic limitations on cell performance; CYCLING WAS ONLY DONE AT 25 CELSIUS AND 45 CELSIUS,
            so any model predictions at low temperature cannot incorporate low temperature degradation modes.
            Discharge capacity

        MODEL SENSITIVITY
            The model predicts degradation rate versus time as a function of temperature and average
            state-of-charge and degradation rate versus equivalent full cycles (charge-throughput) as 
            a function of average state-of-charge during a cycle, depth-of-discharge, and average of the
            charge and discharge C-rates.

        MODEL LIMITATIONS
            There is no influence of TEMPERATURE on CYCLING DEGRADATION RATE due to limited data. This is
            NOT PHYSICALLY REALISTIC AND IS BASED ON LIMITED DATA.
    """

    def __init__(self, degradation_scalar: float = 1, label: str = "LFP-Gr Sony-Murata"):
        # States: Internal states of the battery model
        self.states = {
            'qLoss_LLI_t': np.array([0]),
            'qLoss_LLI_EFC': np.array([0]),
            'qLoss_BreakIn_EFC': np.array([1e-10]),
            'rGain_LLI_t': np.array([0]),
            'rGain_LLI_EFC': np.array([0]),
        }

        # Outputs: Battery properties derived from state values
        self.outputs = {
            'q': np.array([1]),
            'q_LLI_t': np.array([1]),
            'q_LLI_EFC': np.array([1]),
            'q_BreakIn_EFC': np.array([1]),
            'r': np.array([1]),
            'r_LLI_t': np.array([1]),
            'r_LLI_EFC': np.array([1]),
        }

        # Stressors: History of stressors on the battery
        self.stressors = {
            'delta_t_days': np.array([np.nan]), 
            't_days': np.array([0]),
            'delta_efc': np.array([np.nan]), 
            'efc': np.array([0]),
            'TdegK': np.array([np.nan]),
            'soc': np.array([np.nan]), 
            'Ua': np.array([np.nan]), 
            'dod': np.array([np.nan]), 
            'Crate': np.array([np.nan]),
        }

        # Rates: History of stressor-dependent degradation rates
        self.rates = {
            'q1': np.array([np.nan]),
            'q3': np.array([np.nan]),
            'q5': np.array([np.nan]),
            'q7': np.array([np.nan]),
            'r_kcal': np.array([np.nan]),
            'r_kcyc': np.array([np.nan]),
        }

        # Expermental range: details on the range of experimental conditions, i.e.,
        # the range we expect the model to be valid in
        self.experimental_range = {
            'cycling_temperature': [20, 40],
            'dod': [0.8, 1],
            'soc': [0, 1],
            'max_rate_charge': 1,
            'max_rate_discharge': 2,
        }

        # Degradation scalar - scales all state changes by a coefficient
        self._degradation_scalar = degradation_scalar
        # Label for plotting
        self._label = label

    # Nominal capacity
    @property
    def cap(self):
        return 3

    # Define life model parameters
    @property
    def _params_life(self):
        return {
            # Capacity fade parameters
            'q2': 0.000130510034211874,
            'q1_b0': 0.989687151293590,             # CHANGE to modify calendar degradation rate
            'q1_b1': -2881067.56019324,
            'q1_b2': 8742.06309157261,
            'q3_b0': 0.000332850281062177,          
            'q3_b1': 734553185711.369,
            'q3_b2': -2.82161575620780e-06,
            'q3_b3': -3284991315.45121,
            'q3_b4': 0.00127227593657290,
            'q8': 0.00303553871631028,
            'q9': 1.43752162947637,
            'q7_b0': 0.582258029148225,
            'q7_soc_skew': 0.0583128906965484,
            'q7_soc_width': 0.208738181522897,
            'q7_dod_skew': -3.80744333129564,
            'q7_dod_width': 1.16126260428210,
            'q7_dod_growth': 25.4130804598602,
            'q6': 1.12847759334355,
            'q5_b0': -6.81260579372875e-06,         # CHANGE to modify cycling degradation rate
            'q5_b1': 2.59615973160844e-05,
            'q5_b2': 2.11559710307295e-06,
            # Resistance growth parameters
            'k_ref_r_cal': 3.4194e-10,    # CHANGE to modify calendar degradation rate
            'Ea_r_cal': 71827,
            'C_r_cal': -3.3903,
            'D_r_cal': 1.5604,
            'A_r_cyc': -0.002,        # CHANGE to modify cycling degradation rate
            'B_r_cyc': 0.0021,
            'C_r_cyc': 6.8477,
            'D_r_cyc': 0.91882
        }
    
    def update_rates(self, stressors: dict):
        # Calculate and update battery degradation rates based on stressor values
        # Inputs:
        #   stressors (dict): output from extract_stressors

        def _sigmoid(x, alpha, beta, gamma):
            return 2*alpha*(1/2 - 1/(1 + np.exp((beta*x)**gamma)))

        def _skewnormpdf(x, skew, width):
            x_prime = (x-0.5)/width
            return 2 * stats.norm.pdf(x_prime) * stats.norm.cdf(skew * (x_prime))

        # Unpack stressors
        t_secs = stressors["t_secs"]
        delta_t_secs = t_secs[-1] - t_secs[0]
        TdegK = stressors["TdegK"]
        soc = stressors["soc"]
        Ua = stressors["Ua"]
        dod = stressors["dod"]
        Crate = stressors["Crate"]
        
        # Grab parameters
        p = self._params_life

        # Calculate the degradation coefficients
        q1 = np.abs(
        p['q1_b0']
        * np.exp(p['q1_b1']*(1/(TdegK**2))*(Ua**0.5))
        * np.exp(p['q1_b2']*(1/TdegK)*(Ua**0.5))
        )
        q3 = np.abs(
            p['q3_b0']
            * np.exp(p['q3_b1']*(1/(TdegK**4))*(Ua**(1/3)))
            * np.exp(p['q3_b2']*(TdegK**3)*(Ua**(1/4)))
            * np.exp(p['q3_b3']*(1/(TdegK**3))*(Ua**(1/3)))
            * np.exp(p['q3_b4']*(TdegK**2)*(Ua**(1/4)))
            )
        q5 = np.abs(
            p['q5_b0']
            + p['q5_b1']*dod
            + p['q5_b2']*np.exp((dod**2)*(Crate**3))
            )
        q7 = np.abs(
            p['q7_b0']
            * _skewnormpdf(soc, p['q7_soc_skew'], p['q7_soc_width'])
            * _skewnormpdf(dod, p['q7_dod_skew'], p['q7_dod_width'])
            * _sigmoid(dod, 1, p['q7_dod_growth'], 1)
            )
        k_temp_r_cal = (
            p['k_ref_r_cal']
            * np.exp((-p['Ea_r_cal'] / 8.3144) * (1/TdegK - 1/298.15))
        )
        k_soc_r_cal = p['C_r_cal'] * (soc - 0.5)**3 + p['D_r_cal']
        k_Crate_r_cyc = p['A_r_cyc'] * Crate + p['B_r_cyc']
        k_dod_r_cyc = p['C_r_cyc']* (dod - 0.5)**3 + p['D_r_cyc']

        # Calculate time based average of each rate
        q1 = np.trapz(q1, x=t_secs) / delta_t_secs
        q3 = np.trapz(q3, x=t_secs) / delta_t_secs
        # q5 = np.trapz(q5, x=t_secs) / delta_t_secs # no time varying inputs
        q7 = np.trapz(q7, x=t_secs) / delta_t_secs
        k_temp_r_cal = np.trapz(k_temp_r_cal, x=t_secs) / delta_t_secs
        k_soc_r_cal = np.trapz(k_soc_r_cal, x=t_secs) / delta_t_secs
        # k_Crate_r_cyc = np.trapz(k_Crate_r_cyc, x=t_secs) / delta_t_secs # no time varying inputs
        # k_dod_r_cyc = np.trapz(k_dod_r_cyc, x=t_secs) / delta_t_secs # no time varying inputs

        # Store rates
        rates = np.array([q1, q3, q5, q7, k_temp_r_cal * k_soc_r_cal, k_Crate_r_cyc * k_dod_r_cyc])
        for k, v in zip(self.rates.keys(), rates):
            self.rates[k] = np.append(self.rates[k], v)
    
    def update_states(self, stressors: dict):
        # Update the battery states, based both on the degradation state as well as the battery performance
        # at the ambient temperature, T_celsius
        # Inputs:
            #   stressors (dict): output from extract_stressors
            
        # Unpack stressors
        delta_t_days = stressors["delta_t_days"]
        delta_efc = stressors["delta_efc"]
        
        # Grab parameters
        p = self._params_life

        # Grab rates, only keep most recent value
        r = self.rates.copy()
        for k, v in zip(r.keys(), r.values()):
            r[k] = v[-1]

        # Calculate incremental state changes
        states = self.states

        # Capacity
        dq_LLI_t = self._degradation_scalar * self._update_sigmoid_state(states['qLoss_LLI_t'][-1], delta_t_days, r['q1'], p['q2'], r['q3'])
        dq_LLI_EFC = self._degradation_scalar * self._update_power_B_state(states['qLoss_LLI_EFC'][-1], delta_efc, r['q5'], p['q6'])
        if delta_efc / delta_t_days > 3: # only evalaute if more than 3 full cycles per day
            dq_BreakIn_EFC = self._degradation_scalar * self._update_sigmoid_state(states['qLoss_BreakIn_EFC'][-1], delta_efc, r['q7'], p['q8'], p['q9'])
        else:
            dq_BreakIn_EFC = 0

        # Resistance
        dr_LLI_t = self._degradation_scalar * r['r_kcal'] * delta_t_days * 86400
        dr_LLI_EFC = self._degradation_scalar * r['r_kcyc'] * delta_efc / 100

        # Accumulate and store states
        dx = np.array([dq_LLI_t, dq_LLI_EFC, dq_BreakIn_EFC, dr_LLI_t, dr_LLI_EFC])
        for k, v in zip(states.keys(), dx):
            x = self.states[k][-1] + v
            self.states[k] = np.append(self.states[k], x)
    
    def update_outputs(self, stressors):
        # Calculate outputs, based on current battery state
        states = self.states

        # Capacity
        q_LLI_t = 1 - states['qLoss_LLI_t'][-1]
        q_LLI_EFC = 1 - states['qLoss_LLI_EFC'][-1]
        q_BreakIn_EFC = 1 - states['qLoss_BreakIn_EFC'][-1]
        q = 1 - states['qLoss_LLI_t'][-1] - states['qLoss_LLI_EFC'][-1] - states['qLoss_BreakIn_EFC'][-1]
        
        # Resistance
        r_LLI_t = 1 + states['rGain_LLI_t'][-1]
        r_LLI_EFC = 1 + states['rGain_LLI_EFC'][-1]
        r = 1 + states['rGain_LLI_t'][-1] + states['rGain_LLI_EFC'][-1]

        # Assemble output
        out = np.array([q, q_LLI_t, q_LLI_EFC, q_BreakIn_EFC, r, r_LLI_t, r_LLI_EFC])
        # Store results
        for k, v in zip(list(self.outputs.keys()), out):
            self.outputs[k] = np.append(self.outputs[k], v)



# Paul Gasper, NREL
# This model is fit to LG MJ1 cell aging data reported as part of the EU EVERLASTING battery project, report D2.3
# https://everlasting-project.eu/wp-content/uploads/2020/03/EVERLASTING_D2.3_final_20200228.pdf
# Cell tests were reported in early 2020, so likely 2018 or 2019 LG MJ1 cells.
# High energy density 18650s but poor cycle life.

import numpy as np
from blast.models.degradation_model import BatteryDegradationModel


class Nmc811_GrSi_LGMJ1_4Ah_Battery(BatteryDegradationModel):
    """
    Model fit to LG MJ1 cell aging data reported as part of the EU EVERLASTING battery project, report D2.3
    https://everlasting-project.eu/wp-content/uploads/2020/03/EVERLASTING_D2.3_final_20200228.pdf
    Cell tests were reported in early 2020, so likely 2018 or 2019 LG MJ1 cells.
    High energy density 18650s but poor cycle life.

    .. note::
        EXPERIMENTAL AGING DATA SUMMARY:
            Calendar aging varied SOC (10%, 70%, 90%) and temperature.
            Cycle aging varied temperature and C-rates; all DOD is 80% (10%-90%). NO ACCELERATED FADE OBSERVED.
            Relative discharge capacity is reported from measurements recorded at 25 Celsius and C/20 rate.

        MODEL SENSITIVITY
            The model predicts degradation rate versus time as a function of temperature and average
            state-of-charge and degradation rate versus equivalent full cycles (charge-throughput) as 
            a function of C-rate, temperature, and depth-of-discharge (DOD dependence is assumed to be linear, no aging data)

        MODEL LIMITATIONS
            Cycle degradation predictions WILL NOT PREDICT KNEE-POINT due to limited data.
            OPERATION AT HIGH DOD PREDCTIONS ARE LIKELY INACCURATE (it is unclear what voltage window corresponds to SOCs defined in the test data).
            NMC811 is known to degrade quickly at voltages above 4.1 V.
    """

    def __init__(self, degradation_scalar: float = 1, label: str = "NMC811-GrSi LG MJ1"):
        # States: Internal states of the battery model
        self.states = {
            'qLoss_t': np.array([0]),
            'qLoss_EFC': np.array([0]),
        }

        # Outputs: Battery properties derived from state values
        self.outputs = {
            'q': np.array([1]),
            'q_t': np.array([1]),
            'q_EFC': np.array([1]),
        }

        # Stressors: History of stressors on the battery
        self.stressors = {
            'delta_t_days': np.array([np.nan]), 
            't_days': np.array([0]),
            'delta_efc': np.array([np.nan]), 
            'efc': np.array([0]),
            'TdegK': np.array([np.nan]),
            'soc': np.array([np.nan]),
            'dod': np.array([np.nan]),
            'Crate': np.array([np.nan]),
        }

        # Rates: History of stressor-dependent degradation rates
        self.rates = {
            'k_cal': np.array([np.nan]),
            'k_cyc': np.array([np.nan]),
        }

        # Expermental range: details on the range of experimental conditions, i.e.,
        # the range we expect the model to be valid in
        self.experimental_range = {
            'cycling_temperature': [0, 50],
            'dod': [0.2, 0.8],
            'soc': [0.1, 0.9],
            'max_rate_charge': 1,
            'max_rate_discharge': 3,
        }

        # Degradation scalar - scales all state changes by a coefficient
        self._degradation_scalar = degradation_scalar
        # Label for plotting
        self._label = label

    # Nominal capacity
    @property
    def cap(self):
        return 3.5

    # Define life model parameters
    @property
    def _params_life(self):
        return {
            # Capacity fade parameters
            'qcal_A': 0.0353,
            'qcal_B': -1.03e+03,
            'qcal_C': 57.7,
            'qcal_p': 0.743,
            'qcyc_A': 1.77e-07,
            'qcyc_B': 8.08e-13,
            'qcyc_C': 2.21e-07,
            'qcyc_D': 2.25e+03,
            'qcyc_E': 1.14e+04,
            'qcyc_p': 0.695,
        }

    # Battery model
    def update_rates(self, stressors):
        # Calculate and update battery degradation rates based on stressor values
        # Inputs:
        #   stressors (dict): output from extract_stressors

        # Unpack stressors
        t_secs = stressors["t_secs"]
        delta_t_secs = t_secs[-1] - t_secs[0]
        TdegK = stressors["TdegK"]
        soc = stressors["soc"]
        Ua = stressors["Ua"]
        dod = stressors["dod"]
        Crate = stressors["Crate"]
        
        # Grab parameters
        p = self._params_life

        # Calculate the degradation coefficients
        k_cal = p['qcal_A'] * np.exp(p['qcal_B']/TdegK) * np.exp(p['qcal_C']*soc/TdegK)
        k_cyc  = (
            (p['qcyc_A'] + p['qcyc_B']*Crate + p['qcyc_C']*dod)
            * (np.exp(p['qcyc_D']/TdegK) + np.exp(-p['qcyc_E']/TdegK))
        )

        # Calculate time based average of each rate
        k_cal = np.trapz(k_cal, x=t_secs) / delta_t_secs
        k_cyc = np.trapz(k_cyc, x=t_secs) / delta_t_secs

        # Store rates
        rates = np.array([k_cal, k_cyc])
        for k, v in zip(self.rates.keys(), rates):
            self.rates[k] = np.append(self.rates[k], v)
    
    def update_states(self, stressors):
        # Update the battery states, based both on the degradation state as well as the battery performance
        # at the ambient temperature, T_celsius
        # Inputs:
            #   stressors (dict): output from extract_stressors
            
        # Unpack stressors
        delta_t_days = stressors["delta_t_days"]
        delta_efc = stressors["delta_efc"]

        # Grab parameters
        p = self._params_life

        # Grab rates, only keep most recent value
        r = self.rates.copy()
        for k, v in zip(r.keys(), r.values()):
            r[k] = v[-1]

        # Calculate incremental state changes
        states = self.states
        # Capacity
        dq_t = self._degradation_scalar * self._update_power_state(states['qLoss_t'][-1], delta_t_days, r['k_cal'], p['qcal_p'])
        dq_EFC = self._degradation_scalar * self._update_power_state(states['qLoss_EFC'][-1], delta_efc, r['k_cyc'], p['qcyc_p'])

        # Accumulate and store states
        dx = np.array([dq_t, dq_EFC])
        for k, v in zip(states.keys(), dx):
            x = self.states[k][-1] + v
            self.states[k] = np.append(self.states[k], x)
    
    def update_outputs(self, stressors):
        # Calculate outputs, based on current battery state
        states = self.states

        # Capacity
        q_t = 1 - states['qLoss_t'][-1]
        q_EFC = 1 - states['qLoss_EFC'][-1]
        q = 1 - states['qLoss_t'][-1] - states['qLoss_EFC'][-1]

        # Assemble output
        out = np.array([q, q_t, q_EFC])
        # Store results
        for k, v in zip(list(self.outputs.keys()), out):
            self.outputs[k] = np.append(self.outputs[k], v)


import sys, inspect


def available_models():
    """
    Return list of names of available models.
    """
    clsmembers = inspect.getmembers(sys.modules['blast.models'], inspect.isclass)
    return [cls[0] for cls in clsmembers if cls[0] != 'BatteryDegradationModel']


# Paul Gasper, NREL
# This model is fit to data reported by Bank et al from commercial NMC-LTO cells.
# https://doi.org/10.1016/j.jpowsour.2020.228566

import numpy as np
from blast.models.degradation_model import BatteryDegradationModel


class Nmc_Lto_10Ah_Battery(BatteryDegradationModel):
    """
    Model fit to data reported by Bank et al from commercial NMC-LTO cells.
    https://doi.org/10.1016/j.jpowsour.2020.228566

    .. note::
        EXPERIMENTAL AGING DATA SUMMARY:
            Calendar aging varies temperature and SOC. There is almost no calendar aging impact
            at all until 80 Celsius.
            Cycle aging varies temperature, C-rate, and depth-of-discharge.

        MODEL SENSITIVITY
            The model predicts degradation rate versus time as a function of temperature and average
            state-of-charge and degradation rate versus equivalent full cycles (charge-throughput) as 
            a function of C-rate, temperature, and depth-of-discharge (DOD dependence is assumed to be linear, no aging data)

        MODEL LIMITATIONS
            Calendar aging has competition between capacity gain and capacity loss. There is an experimental
            case (80 Celsius, 5% SOC) that has complex behavior not modeled here.
            Astonishingly enough, the cycling degradation model is actually _overestimating_ capacity fade for most cases.
            The exception here is at very high temperature (60+ Celsius), where the fade is high, but not quite as high as observed degradation.
    """

    def __init__(self, degradation_scalar: float = 1, label: str = "NMC-LTO"):
        # States: Internal states of the battery model
        self.states = {
            'qLoss_t': np.array([0]),
            'qGain_t': np.array([0]),
            'qLoss_EFC': np.array([0]),
        }

        # Outputs: Battery properties derived from state values
        self.outputs = {
            'q': np.array([1]),
            'q_t_loss': np.array([1]),
            'q_t_gain': np.array([1]),
            'q_EFC': np.array([1]),
        }

        # Stressors: History of stressors on the battery
        self.stressors = {
            'delta_t_days': np.array([np.nan]), 
            't_days': np.array([0]),
            'delta_efc': np.array([np.nan]), 
            'efc': np.array([0]),
            'TdegK': np.array([np.nan]),
            'soc': np.array([np.nan]),
            'dod': np.array([np.nan]),
            'Crate': np.array([np.nan]),
        }

        # Rates: History of stressor-dependent degradation rates
        self.rates = {
            'alpha': np.array([np.nan]),
            'beta': np.array([np.nan]),
            'gamma': np.array([np.nan]),
        }

        # Expermental range: details on the range of experimental conditions, i.e.,
        # the range we expect the model to be valid in
        self.experimental_range = {
            'cycling_temperature': [30, 60],
            'dod': [0, 1],
            'soc': [0, 1],
            'max_rate_charge': 10,
            'max_rate_discharge': 10,
        }

        # Degradation scalar - scales all state changes by a coefficient
        self._degradation_scalar = degradation_scalar
        # Label for plotting
        self._label = label

    # Nominal capacity
    @property
    def cap(self):
        return 10.2

    # Define life model parameters
    @property
    def _params_life(self):
        return {
            # Capacity fade parameters
            'alpha_0': 3.11e+11,
            'alpha_1': -34.8,
            'alpha_2': 1.07,
            'alpha_p': 0.473,
            'beta_0': 7.86e+10,
            'beta_1': -35.8,
            'beta_2': 3.94,
            'beta_p': -0.553,
            'gamma_0': 1.29,
            'gamma_1': 7.83e-05,
            'gamma_2': 4.02,
            'gamma_3': -8.33,
            'gamma_p': 0.526,
        }
        
    # Battery model
    def update_rates(self, stressors):
        # Calculate and update battery degradation rates based on stressor values
        # Inputs:
        #   stressors (dict): output from extract_stressors

        # Unpack stressors
        t_secs = stressors["t_secs"]
        delta_t_secs = t_secs[-1] - t_secs[0]
        TdegK = stressors["TdegK"]
        soc = stressors["soc"]
        Ua = stressors["Ua"]
        dod = stressors["dod"]
        Crate = stressors["Crate"]
        TdegKN = TdegK / (273.15 + 45)

        # Grab parameters
        p = self._params_life

        # Calculate the degradation coefficients
        alpha = p['alpha_0'] * np.exp(p['alpha_1']/TdegKN) * np.exp(p['alpha_2']*soc/TdegKN)
        beta = p['beta_0'] * np.exp(p['beta_1']/TdegKN) * np.exp(p['beta_2']*soc/TdegKN)
        gamma = (
            (p['gamma_0'] + p['gamma_1']*Crate + p['gamma_2']*(dod**3))
            * np.exp(p['gamma_3']/TdegKN)
        )

        # Calculate time based average of each rate
        alpha = np.trapz(alpha, x=t_secs) / delta_t_secs
        beta = np.trapz(beta, x=t_secs) / delta_t_secs
        gamma = np.trapz(gamma, x=t_secs) / delta_t_secs

        # Store rates
        rates = np.array([alpha, beta, gamma])
        for k, v in zip(self.rates.keys(), rates):
            self.rates[k] = np.append(self.rates[k], v)

    def update_states(self, stressors):
        # Update the battery states, based both on the degradation state as well as the battery performance
        # at the ambient temperature, T_celsius
        # Inputs:
            #   stressors (dict): output from extract_stressors
            
        # Unpack stressors
        delta_t_days = stressors["delta_t_days"]
        delta_efc = stressors["delta_efc"]
        
        # Grab parameters
        p = self._params_life

        # Grab rates, only keep most recent value
        r = self.rates.copy()
        for k, v in zip(r.keys(), r.values()):
            r[k] = v[-1]

        # Calculate incremental state changes
        states = self.states
        # Capacity
        dq_t_gain = self._degradation_scalar * self._update_power_state(states['qGain_t'][-1], delta_t_days, r['alpha'], p['alpha_p'])
        dq_t_loss = self._degradation_scalar * self._update_power_state(states['qLoss_t'][-1], delta_t_days, r['beta'], p['beta_p'])
        dq_EFC = self._degradation_scalar * self._update_power_state(states['qLoss_EFC'][-1], delta_efc, r['gamma'], p['gamma_p'])

        # Accumulate and store states
        dx = np.array([dq_t_loss, dq_t_gain, dq_EFC])
        for k, v in zip(states.keys(), dx):
            x = self.states[k][-1] + v
            self.states[k] = np.append(self.states[k], x)
    
    def update_outputs(self, stressors):
        # Calculate outputs, based on current battery state
        states = self.states

        # Capacity
        q_t_loss = 1 - states['qLoss_t'][-1]
        q_t_gain = 1 + states['qGain_t'][-1]
        q_EFC = 1 - states['qLoss_EFC'][-1]
        q = 1 - states['qLoss_t'][-1] + states['qGain_t'][-1] - states['qLoss_EFC'][-1]

        # Assemble output
        out = np.array([q, q_t_loss, q_t_gain, q_EFC])
        # Store results
        for k, v in zip(list(self.outputs.keys()), out):
            self.outputs[k] = np.append(self.outputs[k], v)


"""Implements rainflow cycle counting algorithm for fatigue analysis according to section 5.4.4 in ASTM E1049-85 (2011)."""

from __future__ import division
from collections import deque, defaultdict
import math


def _get_round_function(ndigits=None):
    if ndigits is None:

        def func(x):
            return x

    else:

        def func(x):
            return round(x, ndigits)

    return func


def reversals(series):
    """Iterate reversal points in the series.
    A reversal point is a point in the series at which the first derivative
    changes sign. Reversal is undefined at the first (last) point because the
    derivative before (after) this point is undefined. The first and the last
    points are treated as reversals.
    Parameters
    ----------
    series : iterable sequence of numbers
    Yields
    ------
    Reversal points as tuples (index, value).
    """
    series = iter(series)

    x_last, x = next(series, None), next(series, None)
    if x_last is None or x is None:
        return

    d_last = x - x_last

    yield 0, x_last
    index = None
    for index, x_next in enumerate(series, start=1):
        if x_next == x:
            continue
        d_next = x_next - x
        if d_last * d_next < 0:
            yield index, x
        x_last, x = x, x_next
        d_last = d_next

    if index is not None:
        yield index + 1, x_next


def extract_cycles(series):
    """Iterate cycles in the series.
    Parameters
    ----------
    series : iterable sequence of numbers
    Yields
    ------
    cycle : tuple
        Each tuple contains (range, mean, count, start index, end index).
        Count equals to 1.0 for full cycles and 0.5 for half cycles.
    """
    points = deque()

    def format_output(point1, point2, count):
        i1, x1 = point1
        i2, x2 = point2
        rng = abs(x1 - x2)
        mean = 0.5 * (x1 + x2)
        return rng, mean, count, i1, i2

    for point in reversals(series):
        points.append(point)

        while len(points) >= 3:
            # Form ranges X and Y from the three most recent points
            x1, x2, x3 = points[-3][1], points[-2][1], points[-1][1]
            X = abs(x3 - x2)
            Y = abs(x2 - x1)

            if X < Y:
                # Read the next point
                break
            elif len(points) == 3:
                # Y contains the starting point
                # Count Y as one-half cycle and discard the first point
                yield format_output(points[0], points[1], 0.5)
                points.popleft()
            else:
                # Count Y as one cycle and discard the peak and the valley of Y
                yield format_output(points[-3], points[-2], 1.0)
                last = points.pop()
                points.pop()
                points.pop()
                points.append(last)
    else:
        # Count the remaining ranges as one-half cycles
        while len(points) > 1:
            yield format_output(points[0], points[1], 0.5)
            points.popleft()


def count_cycles(series, ndigits=None, nbins=None, binsize=None):
    """Count cycles in the series.

    Args:
        series : iterable sequence of numbers
        ndigits (int)   Number of digits to round cycle magnitudes to before counting.
                        Use a negative value to round to tens, hundreds, etc.
        nbins (int)     Number of cycle-counting bins.
        binsize (int)   Width of each cycle-counting bin

        Arguments ndigits, nbins and binsize are mutually exclusive.

    Returns:
        A sorted list containing pairs of range and cycle count.
        The counts may not be whole numbers because the rainflow counting
        algorithm may produce half-cycles. If binning is used then ranges
        correspond to the right (high) edge of a bin.
    """
    if sum(value is not None for value in (ndigits, nbins, binsize)) > 1:
        raise ValueError("Arguments ndigits, nbins and binsize are mutually exclusive")

    counts = defaultdict(float)
    cycles = (
        (rng, count) for rng, mean, count, i_start, i_end in extract_cycles(series)
    )

    if nbins is not None:
        binsize = (max(series) - min(series)) / nbins

    if binsize is not None:
        nmax = 0
        for rng, count in cycles:
            quotient = rng / binsize
            n = int(math.ceil(quotient))  # using int for Python 2 compatibility

            if nbins and n > nbins:
                # Due to floating point accuracy we may get n > nbins,
                # in which case we move rng to the preceeding bin.
                if (quotient % 1) > 1e-6:
                    raise Exception("Unexpected error")
                n = n - 1

            counts[n * binsize] += count
            nmax = max(n, nmax)

        for i in range(1, nmax):
            counts.setdefault(i * binsize, 0.0)

    elif ndigits is not None:
        round_ = _get_round_function(ndigits)
        for rng, count in cycles:
            counts[round_(rng)] += count

    else:
        for rng, count in cycles:
            counts[rng] += count

    return sorted(counts.items())



from .degradation_model import BatteryDegradationModel
from .lfp_gr_250AhPrismatic_2019 import Lfp_Gr_250AhPrismatic
from .lfp_gr_SonyMurata3Ah_2018 import Lfp_Gr_SonyMurata3Ah_Battery
from .lmo_gr_NissanLeaf66Ah_2ndLife_2020 import Lmo_Gr_NissanLeaf66Ah_2ndLife_Battery
from .nca_grsi_SonyMurata2p5Ah_2023 import NCA_GrSi_SonyMurata2p5Ah_Battery
from .nca_gr_Panasonic3Ah_2018 import Nca_Gr_Panasonic3Ah_Battery
from .nmc111_gr_Kokam75Ah_2017 import Nmc111_Gr_Kokam75Ah_Battery
from .nmc111_gr_Sanyo2Ah_2014 import Nmc111_Gr_Sanyo2Ah_Battery
from .nmc811_grSi_LGMJ1_4Ah_2020 import Nmc811_GrSi_LGMJ1_4Ah_Battery
from .nmc_gr_50Ah_B1_2020 import NMC_Gr_50Ah_B1
from .nmc_gr_50Ah_B2_2020 import NMC_Gr_50Ah_B2
from .nmc_gr_75Ah_A_2019 import NMC_Gr_75Ah_A
from .nmc_lto_10Ah_2020 import Nmc_Lto_10Ah_Battery

from ._available_models import available_models

__all__ = ['BatteryDegradationModel', 'Lfp_Gr_250AhPrismatic', 'Lfp_Gr_SonyMurata3Ah_Battery', 'Lmo_Gr_NissanLeaf66Ah_2ndLife_Battery', 'NCA_GrSi_SonyMurata2p5Ah_Battery', 'NMC_Gr_50Ah_B1', 'NMC_Gr_50Ah_B2', 'NMC_Gr_75Ah_A', 'Nca_Gr_Panasonic3Ah_Battery', 'Nmc111_Gr_Kokam75Ah_Battery', 'Nmc111_Gr_Sanyo2Ah_Battery', 'Nmc811_GrSi_LGMJ1_4Ah_Battery', 'Nmc_Lto_10Ah_Battery', '_available_models']



# Paul Gasper, NREL
# This model is fit to SECOND LIFE data on Nissan Leaf half-modules (2p cells) by Braco et al.
# https://doi.org/10.1109/EEEIC/ICPSEUROPE54979.2022.9854784 (calendar aging data)
# https://doi.org/10.1016/j.est.2020.101695 (cycle aging data)
# Note that these cells are already hugely degraded, starting out at an average relative capacity
# of 70%. So the model reports q and qNew, where qNew is relative to initial

import numpy as np
from blast.models.degradation_model import BatteryDegradationModel


class Lmo_Gr_NissanLeaf66Ah_2ndLife_Battery(BatteryDegradationModel):
    """
    Model fit to SECOND LIFE data on Nissan Leaf half-modules (2p cells) by Braco et al.
    https://doi.org/10.1109/EEEIC/ICPSEUROPE54979.2022.9854784 (calendar aging data)
    https://doi.org/10.1016/j.est.2020.101695 (cycle aging data)
    Note that these cells are already hugely degraded, starting out at an average relative capacity
    of 70%. So the model reports q and qNew, where qNew is relative to initial.

    .. note ::
        EXPERIMENTAL AGING DATA SUMMARY:
            Calendar aging widely varied SOC and temperature.
            Cycle aging is only at a single condition (25 Celsius, 100% DOD, 1C-1C).

        MODEL SENSITIVITY
            The model predicts degradation rate versus time as a function of temperature and average
            state-of-charge and degradation rate is only a function equivalent full cycles.

        MODEL LIMITATIONS
            Cycling degradation IS ONLY A FUNCTION OF CHARGE THROUGHPUT due to limited aging data.
            Cycling degradation predictions ARE ONLY VALID NEAR 25 CELSIUS, 100% DOD, 1 C CHARGE/DISCHARGE RATE.
    """

    def __init__(self, degradation_scalar: float = 1, label: str = "LMO-Gr Nissan Leaf"):
        # States: Internal states of the battery model
        self.states = {
            'qLoss_t': np.array([0]),
            'qLoss_EFC': np.array([0]),
        }

        # Outputs: Battery properties derived from state values
        self.outputs = {
            'q': np.array([1]),
            'q_t': np.array([1]),
            'q_EFC': np.array([1]),
            'qNew': np.array([0.7]),
        }

        # Stressors: History of stressors on the battery
        self.stressors = {
            'delta_t_days': np.array([np.nan]), 
            't_days': np.array([0]),
            'delta_efc': np.array([np.nan]), 
            'efc': np.array([0]),
            'TdegK': np.array([np.nan]),
            'soc': np.array([np.nan]),
        }

        # Rates: History of stressor-dependent degradation rates
        self.rates = {
            'k_cal': np.array([np.nan]),
        }

        # Expermental range: details on the range of experimental conditions, i.e.,
        # the range we expect the model to be valid in
        self.experimental_range = {
            'cycling_temperature': [20, 30],
            'dod': [0.8, 1],
            'soc': [0, 1],
            'max_rate_charge': 1,
            'max_rate_discharge': 1,
        }

        # Degradation scalar - scales all state changes by a coefficient
        self._degradation_scalar = degradation_scalar
        # Label for plotting
        self._label = label

    # Nominal capacity
    @property
    def cap_2ndLife(self):
        return 46
    
    @property
    def cap(self):
        return 66

    # Define life model parameters
    @property
    def _params_life(self):
        return {
            # Capacity fade parameters
            'qcal_A': 3.25e+08,
            'qcal_B': -7.58e+03,
            'qcal_C': 162,
            'qcal_p': 0.464,
            'qcyc_A': 7.58e-05,
            'qcyc_p': 1.08,
        }
        
    # Battery model    
    def update_rates(self, stressors):
        # Calculate and update battery degradation rates based on stressor values
        # Inputs:
        #   stressors (dict): output from extract_stressors

        # Unpack stressors
        t_secs = stressors["t_secs"]
        delta_t_secs = t_secs[-1] - t_secs[0]
        TdegK = stressors["TdegK"]
        soc = stressors["soc"]
        Ua = stressors["Ua"]
        dod = stressors["dod"]
        Crate = stressors["Crate"]

        # Grab parameters
        p = self._params_life

        # Calculate the degradation coefficients
        k_cal = p['qcal_A'] * np.exp(p['qcal_B']/TdegK) * np.exp(p['qcal_C']*soc/TdegK)

        # Calculate time based average of each rate
        k_cal = np.trapz(k_cal, x=t_secs) / delta_t_secs

        # Store rates
        rates = np.array([k_cal])
        for k, v in zip(self.rates.keys(), rates):
            self.rates[k] = np.append(self.rates[k], v)

    def update_states(self, stressors):
        # Update the battery states, based both on the degradation state as well as the battery performance
        # at the ambient temperature, T_celsius
        # Inputs:
            #   stressors (dict): output from extract_stressors
            
        # Unpack stressors
        delta_t_days = stressors["delta_t_days"]
        delta_efc = stressors["delta_efc"]
        
        # Grab parameters
        p = self._params_life

        # Grab rates, only keep most recent value
        r = self.rates.copy()
        for k, v in zip(r.keys(), r.values()):
            r[k] = v[-1]

        # Calculate incremental state changes
        states = self.states
        # Capacity
        dq_t = self._degradation_scalar * self._update_power_state(states['qLoss_t'][-1], delta_t_days, r['k_cal'], p['qcal_p'])
        dq_EFC = self._degradation_scalar * self._update_power_state(states['qLoss_EFC'][-1], delta_efc, p['qcyc_A'], p['qcyc_p'])

        # Accumulate and store states
        dx = np.array([dq_t, dq_EFC])
        for k, v in zip(states.keys(), dx):
            x = self.states[k][-1] + v
            self.states[k] = np.append(self.states[k], x)
    
    def update_outputs(self, stressors):
        # Calculate outputs, based on current battery state
        states = self.states

        # Capacity
        q_t = 1 - states['qLoss_t'][-1]
        q_EFC = 1 - states['qLoss_EFC'][-1]
        q = 1 - states['qLoss_t'][-1] - states['qLoss_EFC'][-1]
        qNew = 0.7 * q

        # Assemble output
        out = np.array([q, q_t, q_EFC, qNew])
        # Store results
        for k, v in zip(list(self.outputs.keys()), out):
            self.outputs[k] = np.append(self.outputs[k], v)


# Paul Gasper, NREL
# This model is fit to Panasonic 18650B NCA-Gr cells. High-ish energy density 18650 cells with adequate lifetime.
# Calendar data is reported by Keil et al (https://dx.doi.org/10.1149/2.0411609jes)
# Cycling data is reported by Preger et al (https://doi.org/10.1149/1945-7111/abae37) and
# is available at batteryarchive.com.
# I'm not aware of any study conducting both calendar aging and cycle aging of these cells.

import numpy as np
from blast.models.degradation_model import BatteryDegradationModel


class Nca_Gr_Panasonic3Ah_Battery(BatteryDegradationModel):

    """
    Model fit to Panasonic 18650B NCA-Gr cells. High-ish energy density 18650 cells with
    adequate lifetime.
    Calendar data is reported by Keil et al (https://dx.doi.org/10.1149/2.0411609jes)
    Cycling data is reported by Preger et al (https://doi.org/10.1149/1945-7111/abae37) and
    is available at batteryarchive.com.
    The authors of BLAST-Lite are not aware of any study conducting both calendar aging and
    cycle aging of these cells.

    .. note::
        EXPERIMENTAL AGING DATA SUMMARY:
            Calendar aging widely varied SOC at 25, 40, and 50 Celsius. 300 days max.
            Cycle aging varied temperature and C-rates, and DOD. Some accelerating fade is observed
            at room temperature and high DODs but isn't modeled well here. That's not a huge problem,
            because the modeled lifetime is quite short anyways.

        MODEL SENSITIVITY
            The model predicts degradation rate versus time as a function of temperature and average
            state-of-charge and degradation rate versus equivalent full cycles (charge-throughput) as 
            a function of C-rate, temperature, and depth-of-discharge (DOD dependence is assumed to be linear, no aging data)

        MODEL LIMITATIONS
            Cycle degradation predictions WILL NOT PREDICT KNEE-POINT due to limited data.
            Cycle aging is only modeled at 25, 35, and 45 Celsius, PREDICTIONS OUTSIDE THIS 
            TEMPERATURE RANGE MAY BE OPTIMISTIC.
    """

    def __init__(self, degradation_scalar: float = 1, label: str = "NCA-Gr Panasonic"):
        # States: Internal states of the battery model
        self.states = {
            'qLoss_t': np.array([0]),
            'qLoss_EFC': np.array([0]),
        }

        # Outputs: Battery properties derived from state values
        self.outputs = {
            'q': np.array([1]),
            'q_t': np.array([1]),
            'q_EFC': np.array([1]),
        }

        # Stressors: History of stressors on the battery
        self.stressors = {
            'delta_t_days': np.array([np.nan]), 
            't_days': np.array([0]),
            'delta_efc': np.array([np.nan]), 
            'efc': np.array([0]),
            'TdegK': np.array([np.nan]),
            'soc': np.array([np.nan]),
            'dod': np.array([np.nan]),
            'Crate': np.array([np.nan]),
        }

        # Rates: History of stressor-dependent degradation rates
        self.rates = {
            'k_cal': np.array([np.nan]),
            'k_cyc': np.array([np.nan]),
        }

        # Expermental range: details on the range of experimental conditions, i.e.,
        # the range we expect the model to be valid in
        self.experimental_range = {
            'cycling_temperature': [15, 35],
            'dod': [0.8, 1],
            'soc': [0, 1],
            'max_rate_charge': 0.5,
            'max_rate_discharge': 2,
        }

        # Degradation scalar - scales all state changes by a coefficient
        self._degradation_scalar = degradation_scalar
        # Label for plotting
        self._label = label

    # Nominal capacity
    @property
    def cap(self):
        return 3.2

    # Define life model parameters
    @property
    def _params_life(self):
        return {
            # Capacity fade parameters
            'qcal_A': 75.4,
            'qcal_B': -3.34e+03,
            'qcal_C': 353,
            'qcal_p': 0.512,
            'qcyc_A': 1.86e-06,
            'qcyc_B': 4.74e-11,
            'qcyc_C': 0.000177,
            'qcyc_D': 3.34e-11,
            'qcyc_E': 2.81e-09,
            'qcyc_p': 0.699,
        }
        
    # Battery model
    def update_rates(self, stressors):
        # Calculate and update battery degradation rates based on stressor values
        # Inputs:
        #   stressors (dict): output from extract_stressors

        # Unpack stressors
        t_secs = stressors["t_secs"]
        delta_t_secs = t_secs[-1] - t_secs[0]
        delta_t_days = stressors["delta_t_days"]
        delta_efc = stressors["delta_efc"]
        TdegK = stressors["TdegK"]
        soc = stressors["soc"]
        Ua = stressors["Ua"]
        dod = stressors["dod"]
        Crate = stressors["Crate"]

        # Grab parameters
        p = self._params_life

        # Calculate the degradation coefficients
        k_cal = p['qcal_A'] * np.exp(p['qcal_B']/TdegK) * np.exp(p['qcal_C']*soc/TdegK)
        k_cyc  = (
            (p['qcyc_A'] + p['qcyc_B']*Crate + p['qcyc_C']*dod)
            * (np.exp(p['qcyc_D']/TdegK) + np.exp(-p['qcyc_E']/TdegK))
        )

        # Calculate time based average of each rate
        k_cal = np.trapz(k_cal, x=t_secs) / delta_t_secs
        k_cyc = np.trapz(k_cyc, x=t_secs) / delta_t_secs

        # Store rates
        rates = np.array([k_cal, k_cyc])
        for k, v in zip(self.rates.keys(), rates):
            self.rates[k] = np.append(self.rates[k], v)

    
    def update_states(self, stressors):
        # Update the battery states, based both on the degradation state as well as the battery performance
        # at the ambient temperature, T_celsius
        # Inputs:
            #   stressors (dict): output from extract_stressors
            
        # Unpack stressors
        delta_t_days = stressors["delta_t_days"]
        delta_efc = stressors["delta_efc"]
        
        # Grab parameters
        p = self._params_life

        # Grab rates, only keep most recent value
        r = self.rates.copy()
        for k, v in zip(r.keys(), r.values()):
            r[k] = v[-1]

        # Calculate incremental state changes
        states = self.states
        # Capacity
        dq_t = self._degradation_scalar * self._update_power_state(states['qLoss_t'][-1], delta_t_days, r['k_cal'], p['qcal_p'])
        dq_EFC = self._degradation_scalar * self._update_power_state(states['qLoss_EFC'][-1], delta_efc, r['k_cyc'], p['qcyc_p'])

        # Accumulate and store states
        dx = np.array([dq_t, dq_EFC])
        for k, v in zip(states.keys(), dx):
            x = self.states[k][-1] + v
            self.states[k] = np.append(self.states[k], x)
    
    def update_outputs(self, stressors):
        # Calculate outputs, based on current battery state
        states = self.states

        # Capacity
        q_t = 1 - states['qLoss_t'][-1]
        q_EFC = 1 - states['qLoss_EFC'][-1]
        q = 1 - states['qLoss_t'][-1] - states['qLoss_EFC'][-1]

        # Assemble output
        out = np.array([q, q_t, q_EFC])
        # Store results
        for k, v in zip(list(self.outputs.keys()), out):
            self.outputs[k] = np.append(self.outputs[k], v)


"""Helper functions for fetching and formatting simulation input data."""

import h5pyd
import pandas as pd
from scipy.spatial import cKDTree
from geopy.geocoders import Nominatim
import numpy as np
from numpy import interp
import importlib
from blast.models._available_models import available_models

def simulate_all_models(*args, **kwargs):
    models = available_models()
    cells = []
    for model in models:
        This_Battery_Model = getattr(importlib.import_module("blast.models"), model)
        cell = This_Battery_Model()
        cell.simulate_battery_life(*args, **kwargs)
        cells.append(cell)
    
    return cells

def get_nsrdb_temperature_data(location: str = "Honolulu, Hawaii") -> pd.DataFrame:
    """
    Get temperature time-series data from NSRDB at the nearest coordinates to the
    provided location, and format for battery life simulation.

    Args:
        location (str):      Descriptive location string.

    Returns:
        pd.DataFrame:  Temperature time series at the specified location.
    """

    def nearest_site(tree, lat_coord, lon_coord):
        lat_lon = np.array([lat_coord, lon_coord])
        dist, pos = tree.query(lat_lon)
        return pos

    loc = Nominatim(user_agent="Geopy Library")
    # entering the location name
    getLoc = loc.geocode(location)
    # printing address
    print("Found location: ", getLoc.address)
    # printing latitude and longitude
    lat = getLoc.latitude
    lon = getLoc.longitude
    coords = (lat, lon)

    # Open NSRDB .h5 file
    f = h5pyd.File("/nrel/nsrdb/v3/nsrdb_2018.h5", "r")
    # Get spatial coordinates
    dset_coords = f["coordinates"][...]
    tree = cKDTree(dset_coords)
    pos_idx = nearest_site(tree, coords[0], coords[1])
    print("Input coordiantes: \t {}".format(coords))
    print("Coordinates of nearest point in NSRDB: \t {}".format(dset_coords[pos_idx]))
    # Extract time_index and convert to datetime
    # NOTE: time_index is saved as byte-strings and must be decoded
    time_index = pd.to_datetime(f["time_index"][...].astype(str))
    # Initialize DataFrame to store time-series data
    time_series = pd.DataFrame(index=time_index)
    # Extract variables needed
    for var in ["air_temperature"]:
        # Get dataset
        ds = f[var]
        # Extract scale factor
        scale_factor = ds.attrs["psm_scale_factor"]
        # Extract site 100 and add to DataFrame
        time_series[var] = ds[:, pos_idx] / scale_factor

    time_series = time_series.reset_index()
    time_series["dt"] = time_series["index"] - time_series["index"][0]
    time_series["Time_s"] = time_series["dt"].values / np.timedelta64(1, "s")
    time_series["Temperature_C"] = time_series["air_temperature"]
    time_series = time_series.drop(columns=["dt", "index", "air_temperature"])

    return time_series


def make_inputs_periodic(
    input_timeseries: dict, interp_time_window_hours: float
) -> pd.DataFrame:
    """
    Linearly interpolate the last 'interp_time_window_hours' of the input to ensure periodicity.

    Args:
        input_timeseries (dict):     Dictionary with keys ['Temperature_C', 'SOC', 'Time_s']
                                    and timeseries values.
        interp_time_window_hours (float):    Number of hours to interpolate, from end of timeseries

    Returns:
        dict:    Dictionary with keys ['Temperature_C', 'SOC', 'Time_s'] after interpolation.
    """

    # Check if required keys are in the input_timeseries dictionary
    required_keys = ["Temperature_C", "SOC", "Time_s"]
    for key in required_keys:
        if key not in input_timeseries:
            raise ValueError(
                f"Required key '{key}' is missing in the input_timeseries dictionary."
            )
    # Check if all values are numpy arrays and of the same length
    lengths = [len(input_timeseries[key]) for key in required_keys]
    if len(set(lengths)) != 1:
        raise ValueError(
            "All numpy arrays in input_timeseries must have the same length."
        )

    # Extract arrays from input_timeseries
    t_secs = input_timeseries["Time_s"]
    temperature = input_timeseries["Temperature_C"]
    soc = input_timeseries["SOC"]

    # Convert interp_time_window_hours to seconds
    interp_time_window_secs = interp_time_window_hours * 3600.0

    # Identify the range of indices to interpolate
    end_time = t_secs[-1]
    start_time = end_time - interp_time_window_secs
    idx = np.argmin(np.abs(t_secs - start_time))

    # Interpolate SOC and Temperature_C
    soc_interp = interp(t_secs[idx:], t_secs[idx:], [soc[idx], soc[0]])
    temperature_interp = interp(
        t_secs[idx:], t_secs[idx:], [temperature[idx], temperature[0]]
    )
    # Replace the last part of soc and temperature arrays with interpolated values
    soc[idx:] = soc_interp
    temperature[idx:] = temperature_interp

    # Update the input_timeseries dictionary with interpolated values
    input_timeseries["SOC"] = soc
    input_timeseries["Temperature_C"] = temperature

    return input_timeseries


def assemble_one_year_input(
    soc: pd.DataFrame, climate: pd.DataFrame, time_interval: int = 3600
) -> pd.DataFrame:
    """
    Construct SOC and climate data into a single dataframe,
    linked by timestamp.

    Args:
        soc (pd.DataFrame):      Dataframe with SOC profile and 'Time_s' timestamp
        climate (pd.DataFrame):  Dataframe with temperature data and 'Time_s' timestamp column
        time_interval (int):     Interval at which the final data is resampled to

    Return:
        pd.DataFrame:  Combined DataFrame with time, soc, and temperature columns.
    """
    # Check if 'Time_s' column exists in both data frames
    if "Time_s" not in soc.columns:
        raise ValueError("soc dataframe does not contain 'Time_s' column.")
    if "Time_s" not in climate.columns:
        raise ValueError("climate dataframe does not contain 'Time_s' column.")
    # Make each data frame one year long
    soc = tile_to_one_year(soc)
    climate = tile_to_one_year(climate)
    # Merge data frames on 'Time_s' column
    combined_df = pd.merge(soc, climate, on="Time_s", how="outer")
    # Sort values by 'Time_s'
    combined_df = combined_df.sort_values("Time_s")
    # Interpolate to fill missing values
    combined_df = combined_df.interpolate(method="linear")
    # Resample to at the input interval (default 1 hour) from 0 to 31104000 (one year)
    index = pd.Index(np.arange(0, 31104000, time_interval), name="Time_s")
    combined_df = combined_df.set_index("Time_s").reindex(index)

    return combined_df.reset_index()


def tile_to_one_year(df: pd.DataFrame) -> pd.DataFrame:
    """
    Take a dataframe with timestamp column called 'Time_s'
    and resample to 1 year.

    Args:
        df (pd.DataFrame):   Dataframe to cut off to 1 year.

    Returns:
        pd.DataFrame:   Dataframe after cutting off to 1 year.
    """
    t_secs = df["Time_s"].values
    if t_secs[-1] < 365 * 24 * 3600:
        df = df.drop(columns="Time_s")
        df_values = df.to_numpy()
        # Tile the inputs. When tiling time, assume the timestep
        # between repeats is the most common timestep.
        n_tile = np.ceil((365 * 24 * 3600) / t_secs[-1]).astype(int)
        delta_t_secs = np.ediff1d(t_secs, to_begin=0)
        delta_t_secs[0] = np.median(delta_t_secs)
        delta_t_secs = np.tile(delta_t_secs, n_tile)
        delta_t_secs[0] = 0
        t_secs = np.cumsum(delta_t_secs)
        df_values = np.tile(df_values, (n_tile, 1))

        # Cutoff once simulation is at simulation_years
        idx_end = np.argwhere(t_secs > 365 * 24 * 3600)
        idx_end = idx_end[0][0]
        t_secs = t_secs[:idx_end]
        df_values = df_values[:idx_end]

        # remake into dataframe
        df = pd.DataFrame(df_values, columns=df.columns)
        df["Time_s"] = t_secs

    return df
